#include <utility/string.h>
#include <utility/memory.h>
#include <memory/memory.h>
#include <i386/manage.h>
#include <console/console.h>
#include <interrupt/irq.h>
#include <interrupt/exception.h>
#include <linker/linker.h>
#include <debugger/breakpoint.h>

static ExceptionHandler breakPointExceptionHandler;
static List breakPoints;

static void
breakpoint_exception_handler (ExceptionNumber number, ExceptionParameter parameter, void * data)
{
     IRQRegisters * registers;
     BreakPoint * breakPoint;
  
     registers = (IRQRegisters *) (I386_STACK_EXCEPTION_OFFSET + 
                                      I386_STACK_EXCEPTION_LIMIT - 
                                      sizeof (IRQRegisters)); 

     -- registers -> EIP;

     breakPoint = breakpoint_find_address ((uint8 *) registers -> EIP);

     if (breakPoint == NULL)
     { 
        console_print_format (NULL, "Unknown breakpoint at address %x!\n", registers -> EIP);

        return;
     }

     if (breakPoint -> isActivated == TRUE)
       return;

     breakPoint -> isActivated = TRUE;

     console_print_format (NULL,
                           "Breakpoint %s:%s at 0x%x activated!\n",
                           breakPoint -> function -> section -> component -> name,
                           breakPoint -> function -> name,
                           breakPoint -> function -> section -> data + breakPoint -> function -> offset);
}

void
breakpoint_disable (BreakPoint * breakPoint)
{
     breakPoint -> isActivated = FALSE;

     * breakPoint -> address = * breakPoint -> code;
}
                  
void
breakpoint_enable (BreakPoint * breakPoint)
{
     breakPoint -> isActivated = FALSE;
        
     * breakPoint -> address = 0xCC;
}

Bool
__breakpoint_enter__ (void)
{
     list_clear (& breakPoints);

     exception_handler_create (& breakPointExceptionHandler,
                               3,
                               breakpoint_exception_handler,
                               NULL,
                               0,
                               "breakpoint");

     return TRUE;
}

void
__breakpoint_exit__ (void)
{
     exception_handler_destroy (& breakPointExceptionHandler);

     while (list_empty (& breakPoints) == FALSE)
       breakpoint_destroy ((BreakPoint *) list_front (& breakPoints));
}

Bool
breakpoint_create (BreakPoint ** result,
                   const char * componentName,
                   const char * functionName)
{
     Component * component = linker_find_component (componentName);
     BreakPoint * breakPoint;
     Export * function;

     if (component == NULL)
       return FALSE;
     
     function = component_find_export (component, functionName);

     if (function == NULL ||
         function -> type != EXPORT_TYPE_FUNCTION)
       return FALSE;

     breakPoint = (BreakPoint *) memory_allocate (sizeof (BreakPoint));

     if (breakPoint == NULL)
       return FALSE;

     breakPoint -> function    = function;
     breakPoint -> address     = function -> section -> data + function -> offset;
     breakPoint -> isActivated = FALSE;
     
     memory_copy (breakPoint -> code,
                  breakPoint -> address,
                  sizeof (breakPoint -> code));

     * breakPoint -> address = 0xCC;

     list_insert_front (& breakPoints, breakPoint);

     if (result != NULL)
       * result = breakPoint;

     return TRUE;
}

void
breakpoint_destroy (BreakPoint * breakPoint)
{
     list_remove (& breakPoint -> breakPointList);

     memory_copy (breakPoint -> address,
                  breakPoint -> code,
                  sizeof (breakPoint -> code));

     memory_free (breakPoint);
}

BreakPoint *
breakpoint_find_address (uint8 * address)
{
     ListIterator position;

     for (position = list_begin (& breakPoints);
          position != list_end (& breakPoints);
          position = list_next (position))
     {
        BreakPoint * breakPoint = (BreakPoint *) position;

        if (address >= breakPoint -> address &&
            address < & breakPoint -> address [sizeof (breakPoint -> code)])
          return breakPoint;
     }

     return NULL;
}

BreakPoint *
breakpoint_find_function (const char * componentName, const char * functionName)
{
     ListIterator position;

     for (position = list_begin (& breakPoints);
          position != list_end (& breakPoints);
          position = list_next (position))
     {
        BreakPoint * breakPoint = (BreakPoint *) position;

        if (string_equal (breakPoint -> function -> section -> component -> name, componentName) == TRUE &&
            string_equal (breakPoint -> function -> name, functionName) == TRUE)
          return breakPoint;
     }

     return NULL;
}

