#include <utility/string.h>
#include <utility/memory.h>
#include <console/console.h>
#include <scheduler/thread.h>
#include <libc/stdlib.h>
#include <libc/string.h>
#include <libc/stdio.h>
#include <libc/ctype.h>
#include <i386/fpu.h>
#include <i386/manage.h>
#include <interrupt/irq.h>
#include <interrupt/exception.h>
#include <debugger/command.h>

static Thread * debuggerThread = NULL;

static Command commands [] =
{
    { "help", command_help },
    { "pong", command_pong },
    { "components", command_components },
    { "component", command_component },
    { "section", command_section },
    { "export", command_export },
    { "break", command_break },
    { "enable", command_enable },
    { "disable", command_disable },
    { "run", command_run }
};

static ExceptionHandler exceptionHandlers [20];

static const char * exceptionNames [20] =
{
    "divide error",
    "debug trap",
    "non-maskable interrupt",
    "breakpoint exception",
    "overflow exception",
    "bound range exceeded exception",
    "invalid opcode exception",
    "device not available exception",
    "double fault exception",
    "coprocessor segment overrun",
    "invalid tss exception",
    "segment not present",
    "stack fault",
    "general protection exception",
    "page fault",
    "reserved",
    "floating point error",
    "alignment check exception",
    "machine check exception",
    "simd floating point exception",
};
    
static void
exceptionHandlerFunction (ExceptionNumber number,
                          ExceptionParameter parameter,
                          void * data)
{
    IRQRegisters * registers = (IRQRegisters *) (I386_STACK_EXCEPTION_OFFSET +
                                                 I386_STACK_EXCEPTION_LIMIT) - 1;

    /* Interrupt any pong activity... */

    console_activate (console_current ());

    console_print_format (NULL, 
                          "exception:\n"
                          "  name: %s\n"
                          "  vector: %u, 0x%x\n"
                          "  parameter: %u, 0x%x\n" 
                          "  EAX: 0x%x, ECX: 0x%x, EDX: 0x%x, EBX: 0x%x\n"
                          "  ESP: 0x%x, EBP: 0x%x, ESI: 0x%x, EDI: 0x%x\n"
                          "  EIP: 0x%x, CS: 0x%x, EFLAGS: 0x%x, TR: 0x%x\n"
                          "  CR0: 0x%x, CR2: 0x%x, CR3: 0x%x, CR4: 0x%x\n\n",
                          exceptionNames [number],
                          number,
                          number,
                          parameter,
                          parameter,
                          registers -> EAX,
                          registers -> ECX,
                          registers -> EDX,
                          registers -> EBX,
                          registers -> ESP,
                          registers -> EBP,
                          registers -> ESI,
                          registers -> EDI,
                          registers -> EIP,
                          registers -> CS,
                          registers -> EFLAGS,
                          i386_get_TR (),
                          i386_get_CR0 (),
                          i386_get_CR2 (),
                          i386_get_CR3 (),
                          i386_get_CR4 ());
}

static void
debuggerLoop (void * data)
{
    static char inputBuffer [4096];

    i386_fpu_set_control_word (I386_FPU_CONTROL_MASK_EXCEPTIONS | i386_fpu_get_control_word ());

    command_help (0, NULL);

    for (;;)
    {
       char * commandName,
            * arguments [16],
            ** argumentCurrent;
       Command * command;

       console_print (NULL, "\ndebugger> ");

       commandName = fgets (inputBuffer, sizeof (inputBuffer), stdin);

       if (commandName == NULL)
         continue;

       while (isspace (* commandName) != 0)
          ++ commandName;

       if (* commandName == '\0')
         continue;

       memory_zero (arguments, sizeof (arguments));

       arguments [0] = commandName;

       for (argumentCurrent = & arguments [1];
            argumentCurrent < & arguments [sizeof (arguments) / sizeof (char *)];
            ++ argumentCurrent)
       {
          char * nextArgument = argumentCurrent [-1];
      
          while (* nextArgument > ' ')
             ++ nextArgument;

          * nextArgument ++ = '\0';
         
          if (nextArgument > & inputBuffer [sizeof (inputBuffer)] ||
              argumentCurrent >= & arguments [sizeof (arguments) / sizeof (char *) - 1])
            break;

          while (isspace (* nextArgument) != 0)
             ++ nextArgument;

          if (* nextArgument == '\0' )
            break;
       
          * argumentCurrent = nextArgument;
       }
      
       for (command = commands;
            command < & commands [sizeof (commands) / sizeof (Command)];
            ++ command)
       {
          if (string_equal (command -> name, commandName) == TRUE)
            break;
       }
       
       if (command >= & commands [sizeof (commands) / sizeof (Command)])
       {
          console_print_format (NULL, "Unknown command %s!\n", commandName);
          
          continue;
       }

       if ((* command -> function) ((Size) (argumentCurrent - & arguments [1]), 
                                    & arguments [1]) == FALSE)
       {
          console_print_format (NULL, "Syntax error in command %s!\n", commandName);

          continue;
       }
    }
}
       
Bool
__debugger_enter__ (void)
{
    ExceptionHandler * handler;
    
    for (handler = exceptionHandlers;
         handler < & exceptionHandlers [sizeof (exceptionHandlers) / sizeof (ExceptionHandler)];
         ++ handler)
    {
       switch (handler - exceptionHandlers)
       {
       case 7:
       case 3:
       case 1:
          continue;
       }
            
       exception_handler_create (handler,
                                 handler - exceptionHandlers,
                                 exceptionHandlerFunction,
                                 NULL,
                                 0,
                                 "debugger");
    }
    
    if (thread_create (& debuggerThread,
                       THREAD_PRIORITY_EXCLUSIVE,
                       THREAD_PRIORITY_EXCLUSIVE,
                       debuggerLoop,
                       NULL,
                       NULL,
                       0) == FALSE)
      return FALSE;

    return TRUE;
}

void
__debugger_exit__ (void)
{
   ExceptionHandler * handler;

   if (debuggerThread != NULL)
     thread_destroy (debuggerThread);

   for (handler = exceptionHandlers;
        handler < & exceptionHandlers [sizeof (exceptionHandlers) / sizeof (ExceptionHandler)];
        ++ handler)
   {
      switch  (handler - exceptionHandlers)
      {
      case 7:
      case 3:
      case 1:
         continue;
      }

      exception_handler_destroy (handler);
   }
}

