#include <scheduler/mutex.h>
#include <memory/memory.h>
#include <i386/io.h>
#include <dma/dma.h>

static Mutex dmaMutex;
static DMAChannel dmaChannels [DMA_CHANNEL_COUNT] =
{
   {
     FALSE, 
     DMA_REGISTER_MASK_8,
     DMA_REGISTER_MODE_8,
     DMA_REGISTER_CLEAR_8,
     DMA_REGISTER_CHANNEL_0_PAGE,
     DMA_REGISTER_CHANNEL_0_ADDRESS,
     DMA_REGISTER_CHANNEL_0_COUNT,
     DMA_COMMAND_CHANNEL_0,
     DMA_CHANNEL_ATTRIBUTE_8 
   },
   {
     FALSE,
     DMA_REGISTER_MASK_8,
     DMA_REGISTER_MODE_8,
     DMA_REGISTER_CLEAR_8,
     DMA_REGISTER_CHANNEL_1_PAGE,
     DMA_REGISTER_CHANNEL_1_ADDRESS,
     DMA_REGISTER_CHANNEL_1_COUNT,
     DMA_COMMAND_CHANNEL_1,
     DMA_CHANNEL_ATTRIBUTE_8
   },
   {
     FALSE,
     DMA_REGISTER_MASK_8,
     DMA_REGISTER_MODE_8,
     DMA_REGISTER_CLEAR_8,
     DMA_REGISTER_CHANNEL_2_PAGE,
     DMA_REGISTER_CHANNEL_2_ADDRESS,
     DMA_REGISTER_CHANNEL_2_COUNT,
     DMA_COMMAND_CHANNEL_2,
     DMA_CHANNEL_ATTRIBUTE_8
   },
   {
     FALSE,
     DMA_REGISTER_MASK_8,
     DMA_REGISTER_MODE_8,
     DMA_REGISTER_CLEAR_8,
     DMA_REGISTER_CHANNEL_3_PAGE,
     DMA_REGISTER_CHANNEL_3_ADDRESS,
     DMA_REGISTER_CHANNEL_3_COUNT,
     DMA_COMMAND_CHANNEL_3,
     DMA_CHANNEL_ATTRIBUTE_8
   },
   {
     FALSE,
     DMA_REGISTER_MASK_16,
     DMA_REGISTER_MODE_16,
     DMA_REGISTER_CLEAR_16,
     DMA_REGISTER_CHANNEL_4_PAGE,
     DMA_REGISTER_CHANNEL_4_ADDRESS,
     DMA_REGISTER_CHANNEL_4_COUNT,
     DMA_COMMAND_CHANNEL_4,
     DMA_CHANNEL_ATTRIBUTE_16
   },
   {
     FALSE,
     DMA_REGISTER_MASK_16,
     DMA_REGISTER_MODE_16,
     DMA_REGISTER_CLEAR_16,
     DMA_REGISTER_CHANNEL_5_PAGE,
     DMA_REGISTER_CHANNEL_5_ADDRESS,
     DMA_REGISTER_CHANNEL_5_COUNT,
     DMA_COMMAND_CHANNEL_5,
     DMA_CHANNEL_ATTRIBUTE_16
   },
   {
     FALSE,
     DMA_REGISTER_MASK_16,
     DMA_REGISTER_MODE_16,
     DMA_REGISTER_CLEAR_16,
     DMA_REGISTER_CHANNEL_6_PAGE,
     DMA_REGISTER_CHANNEL_6_ADDRESS,
     DMA_REGISTER_CHANNEL_6_COUNT,
     DMA_COMMAND_CHANNEL_6,
     DMA_CHANNEL_ATTRIBUTE_16
   },
   {
     FALSE,
     DMA_REGISTER_MASK_16,
     DMA_REGISTER_MODE_16,
     DMA_REGISTER_CLEAR_16,
     DMA_REGISTER_CHANNEL_7_PAGE,
     DMA_REGISTER_CHANNEL_7_ADDRESS,
     DMA_REGISTER_CHANNEL_7_COUNT,
     DMA_COMMAND_CHANNEL_7,
     DMA_CHANNEL_ATTRIBUTE_16
   }
};

Bool
__dma_enter__ (void)
{
    return mutex_create (& dmaMutex);
}

void
__dma_exit__ (void)
{
    mutex_destroy (& dmaMutex);
}

const DMAChannel *
dma_channel_allocate (DMAChannelNumber number, DMAChannelAttributes attributesWanted)
{
    DMAChannel * channel;

    mutex_lock (& dmaMutex);

    if (number < sizeof (dmaChannels) / sizeof (DMAChannel))
    {
       channel = & dmaChannels [number];

       if (channel -> isAllocated == TRUE)
       {
	  mutex_unlock (& dmaMutex);

	  return NULL;
       }
    }
    else
    {
       for (channel = dmaChannels;
            channel < & dmaChannels [sizeof (dmaChannels) / sizeof (DMAChannel)];
            ++ channel)
         if ((channel -> attributes & attributesWanted) != 0 &&
             channel -> isAllocated == FALSE)
           break;
       
       if (channel >= & dmaChannels [sizeof (dmaChannels) / sizeof (DMAChannel)])
       {
         mutex_unlock (& dmaMutex);
         
         return NULL;
       }
    }

    channel -> isAllocated = TRUE;

    mutex_unlock (& dmaMutex);

    return channel;
}

void
dma_channel_free (const DMAChannel * channel)
{
    ((DMAChannel *) channel) -> isAllocated = FALSE;
}

void *
dma_buffer_allocate (Size size)
{
    if (size >= DMA_COUNT_LIMIT)
      return NULL;

    return memory_allocate_within ((void *) 0,
                                   (void *) (DMA_ADDRESS_LIMIT - DMA_COUNT_LIMIT),
                                   size);
}

void
dma_buffer_free (void * buffer)
{
    memory_free (buffer);
}

Bool
dma_transfer_start (const DMAChannel * channel, DMATransfer transfer, void * address, Size count)
{
    if ((Address) address >= DMA_ADDRESS_LIMIT ||
        ((Address) address % DMA_COUNT_LIMIT) + count > DMA_COUNT_LIMIT)
      return FALSE;

    i386_output_8 (channel -> maskRegister, DMA_COMMAND_MASK_SET | channel -> commandModifier);

    i386_output_8 (channel -> clearRegister, DMA_COMMAND_CLEAR);
    
    i386_output_8 (channel -> modeRegister, transfer |
                                            DMA_COMMAND_MODE_INCREMENT |
                                            DMA_COMMAND_MODE_SINGLE |
                                            DMA_COMMAND_MODE_AUTO |
                                            channel -> commandModifier);
    
    i386_output_8 (channel -> addressRegister, (Address) address & 0xFF);
    i386_output_8 (channel -> addressRegister, ((Address) address >> 8) & 0xFF);
    i386_output_8 (channel -> pageRegister, ((Address) address >> 16) & 0xFF);

    i386_output_8 (channel -> countRegister, count & 0xFF);
    i386_output_8 (channel -> countRegister, (count >> 8) & 0xFF);

    i386_output_8 (channel -> maskRegister, DMA_COMMAND_MASK_CLEAR | channel -> commandModifier);

    return TRUE;
}

void
dma_transfer_stop (const DMAChannel * channel)
{
    i386_output_8 (channel -> maskRegister, DMA_COMMAND_MASK_SET | channel -> commandModifier);

    i386_output_8 (channel -> clearRegister, DMA_COMMAND_CLEAR);

    i386_output_8 (channel -> maskRegister, DMA_COMMAND_MASK_CLEAR | channel -> commandModifier);
}


