#include <utility/types.h>
#include <utility/memory.h>
#include <scheduler/thread.h>
#include <page/allocator.h>
#include <memory/memory.h>
#include <console/console.h>
#include <i386/manage.h>
#include <i386/page.h>
#include <i386/boot.h>
#include <linker/interface.h>
#include <linker/component.h>
#include <linker/linker.h>
#include <xcom/xcom.h>

static void
bootComponents (void * data)
{
    uint8 * image = (uint8 *) boot_information.loadAddress,
          * currentXCom;
    XComHeader * header = (XComHeader *) image;
    Component * component;
    Interface * interface;
    ListIterator position;

    console_print_format (NULL,
                          "POS (%s)\n"
                          "Loading boot components:\n",
                          __DATE__);

    console_print_format (NULL, "  * [address=0x%x, size=%u] boot... loaded.\n", image, header -> fileSize);
    
    component = linker_find_component ("boot");

    if (component != NULL)
    for (position = list_begin (& component -> interfaces);
         position != list_end (& component -> interfaces);
         position = list_next (position))
    {
         interface = (Interface *) position;

         console_print_format (NULL, "      + %s\n", interface -> name);
    }
 
    for (currentXCom = & image [header -> fileSize];
         currentXCom < & image [boot_information.imageSize];
         currentXCom += header -> fileSize)
    {
        ListIterator position;

        header = (XComHeader *) currentXCom;

        console_print_format (NULL, "  + [address=0x%x, size=%u] ", currentXCom, header -> fileSize);

        if (xcom_load (& component, currentXCom, header -> fileSize) == FALSE)
        {
           console_print (NULL, "failed!\n");

           continue;
        }

        console_print_format (NULL, "%s... ", component -> name);

        if (component_resolve_imports (component) == FALSE)
        {
           component_destroy (component);

           console_print (NULL, "unresolved imports!\n");

           continue;
        }

        component_relocate_sections (component);

        console_print_format (NULL, "loaded.\n");

        if (component_call_constructors (component) == FALSE)
        {
           component_destroy (component);

           console_print (NULL, "      - constructor failed!\n");
           
           continue;
        }

        for (position = list_begin (& component -> interfaces);
             position != list_end (& component -> interfaces);
             position = list_next (position))
        {
            interface = (Interface *) position;

            console_print_format (NULL, "      + %s\n", interface -> name);
        }
    }

    console_print (NULL, "\n");
}

Bool
__boot_enter__ (void)
{
    return thread_create (NULL,
                          THREAD_PRIORITY_EXCLUSIVE,
                          THREAD_PRIORITY_EXCLUSIVE,
                          bootComponents,
                          NULL,
                          NULL,
                          0);
}

Bool
__i386_enter__ (void)
{
    Size lowSize = (0xA0000 - I386_MEMORY_OFFSET) >> I386_PAGE_ALIGNMENT,
         highSize = boot_information.memorySize - (0x100000 >> I386_PAGE_ALIGNMENT),
	 lowAllocatorSize,
	 highAllocatorSize,
	 lowAlignment,
	 highAlignment;
    PageAllocator * lowAllocator,
                  * highAllocator;

    if (memory_compare (I386_BOOT_MAGIC, boot_information.magic, sizeof (boot_information.magic)) != COMPARE_EQUAL)
      return FALSE;

    console_print_format (NULL,
                          "Boot information:\n"
	   	          "  * magic: %s\n"
		          "  * image size: %d B\n"
			  "  * load address: 0x%x\n"
			  "  * memory size: %d KB\n\n",
			  boot_information.magic,
			  boot_information.imageSize,
			  boot_information.loadAddress,
			  boot_information.memorySize * I386_PAGE_SIZE / 1024);

    highAlignment     = memory_alignment (0, 32 - I386_PAGE_ALIGNMENT, highSize);
    highAllocatorSize = page_allocator_size (0x100000 >> I386_PAGE_ALIGNMENT,
                                             boot_information.memorySize,
                                             highAlignment);
    highAllocator     = (PageAllocator *) (boot_information.loadAddress + boot_information.imageSize);

    if (page_allocator_create (highAllocator,
                               0x100000 >> I386_PAGE_ALIGNMENT,
                               boot_information.memorySize,
                               highAlignment, 1, 0) == FALSE)
      return FALSE;

    page_allocator_insert (highAllocator);

    lowAlignment     = memory_alignment (0, 32 - I386_PAGE_ALIGNMENT, lowSize);
    lowAllocatorSize = page_allocator_size (I386_MEMORY_OFFSET >> I386_PAGE_ALIGNMENT,
		                            0xA0000 >> I386_PAGE_ALIGNMENT,
					    lowAlignment);
    lowAllocator     = (PageAllocator *) (boot_information.loadAddress + boot_information.imageSize + highAllocatorSize);

    if (page_allocator_create (lowAllocator,
		               I386_MEMORY_OFFSET >> I386_PAGE_ALIGNMENT,
			       0xA0000 >> I386_PAGE_ALIGNMENT,
			       lowAlignment, 0, 0) == FALSE)
      return FALSE;
    
    page_allocator_insert (lowAllocator);

    memory_claim ((void *) boot_information.loadAddress,
		  boot_information.imageSize + highAllocatorSize + lowAllocatorSize);

    return TRUE;
}

