#ifndef __I386_FPU_H__
#define __I386_FPU_H__

#include <utility/types.h>
#include <utility/inline.h>

typedef enum
{
        I386_FPU_STATE_INITIALIZE = 0,
        I386_FPU_STATE_SAVE       = 1,
        I386_FPU_STATE_RESTORE    = 2
} i386FPUState;

typedef struct
{
	uint32	     control;
	uint32	     status;
	uint32	     tag;
	uint32	     eip;
	uint32	     cs;
	uint32	     opCode;
	uint32	     data;
	uint32	     ds;
	uint8	     registers [10 * 8];
        i386FPUState state;
} i386FPUContext;

typedef enum
{
   I386_FPU_STATUS_INVALID_OPERATION     = (1 << 0),
   I386_FPU_STATUS_DENORMALIZED_OPERATND = (1 << 1),
   I386_FPU_STATUS_ZERO_DIVIDE_EXCEPTION = (1 << 2),
   I386_FPU_STATUS_OVERFLOW_EXCEPTION    = (1 << 3),
   I386_FPU_STATUS_UNDERFLOW_EXCEPTION   = (1 << 4),
   I386_FPU_STATUS_PRECISION_EXCEPTION   = (1 << 5),
   I386_FPU_STATUS_STACK_FAULT           = (1 << 6),
   I386_FPU_STATUS_ERROR_SUMMARY         = (1 << 7),
   I386_FPU_STATUS_CONDITION_CODE_0      = (1 << 8),
   I386_FPU_STATUS_CONDITION_CODE_1      = (1 << 9),
   I386_FPU_STATUS_CONDITION_CODE_2      = (1 << 10),
   I386_FPU_STATUS_TOP_OF_STACK          = (3 << 11),
   I386_FPU_STATUS_CONDITION_CODE_3      = (1 << 14),
   I386_FPU_STATUS_BUSY                  = (1 << 15)
} i386FPUStatus;

typedef enum
{
   I386_FPU_CONTROL_MASK_INVALID_OPERATION     = (1 << 0),
   I386_FPU_CONTROL_MASK_DENORMALIZED_OPERAND  = (1 << 1),
   I386_FPU_CONTROL_MASK_ZERO_DIVIDE_EXCEPTION = (1 << 2),
   I386_FPU_CONTROL_MASK_OVERFLOW_EXCEPTION    = (1 << 3),
   I386_FPU_CONTROL_MASK_UNDERFLOW_EXCEPTION   = (1 << 4),
   I386_FPU_CONTROL_MASK_PRECISION_EXCEPTION   = (1 << 5),
   I386_FPU_CONTROL_MASK_EXCEPTIONS            = (0x3F << 0),
   I386_FPU_CONTROL_PRECISION_TYPE             = (3 << 8),
   I386_FPU_CONTROL_ROUNDING_TYPE              = (3 << 10),
   I386_FPU_CONTROL_INFINITY_CONTROL           = (1 << 12)
} i386FPUControl;

typedef enum
{
   I386_FPU_PRECISION_SINGLE   = (0 << 8),
   I386_FPU_PRECISION_DOUBLE   = (2 << 8),
   I386_FPU_PRECISION_EXTENDED = (3 << 8)
} i386FPUPrecision;

typedef enum
{
   I386_FPU_ROUNDING_NEAREST   = (0 << 10),
   I386_FPU_ROUNDING_DOWN      = (1 << 10),
   I386_FPU_ROUNDING_UP        = (2 << 10),
   I386_FPU_ROUNDING_TRUNCATE  = (3 << 10)
} i386FPURounding;

typedef enum
{
   I386_FPU_TAG_VALID   = 0,
   I386_FPU_TAG_ZERO    = 1,
   I386_FPU_TAG_SPECIAL = 2,
   I386_FPU_TAG_EMPTY   = 3
} i386FPUTag;

#ifdef __GNUC__
#include <utility/assembly.h>

INLINE uint16	i386_fpu_get_control_word	(void)
{
   uint16 value;

   ASSEMBLY ("fnstcw %0" : "=m" (value) :);

   return value;
}

INLINE void	i386_fpu_set_control_word	(uint16 value)
{ ASSEMBLY ("fldcw %0" : : "m" (value)); }

INLINE uint16	i386_fpu_get_status_word	(void)
{ 
   uint16 value;

   ASSEMBLY ("fnstsw %0" : "=r" (value) :);

   return value;
}

INLINE void	i386_fpu_save			(i386FPUContext * context)
{ ASSEMBLY ("fnsave %0" : "=m" (* context) :); }

INLINE void	i386_fpu_restore		(const i386FPUContext * context)
{ ASSEMBLY ("frstor %0" : : "m" (* context)); }

INLINE void	i386_fpu_initialize		(void)
{ ASSEMBLY ("fninit"); }

INLINE void	i386_fpu_wait			(void)
{ ASSEMBLY ("fwait"); }

#else /* ! __GNUC__ */

extern uint16   i386_fpu_get_control_word       (void);
extern void     i386_fpu_set_control_word       (uint16);
extern uint16   i386_fpu_get_status_word        (void);

extern void     i386_fpu_save                   (i386FPUContext *);
extern void     i386_fpu_restore                (const i386FPUContext *);

extern void     i386_fpu_initialize             (void);
extern void     i386_fpu_wait                   (void);

#endif /* __GNUC__ */

#endif /* __I386_FPU_H__ */

