#ifndef __IDE_CONTROLLER_H__
#define __IDE_CONTROLLER_H__

#include <i386/io.h>
#include <utility/inline.h>
#include <scheduler/semaphore.h>
#include <timer/timer.h>
#include <interrupt/interrupt.h>
#include <ide/ide.h>
#include <ide/drive.h>

typedef enum
{
        IDE_CONTROLLER_IO_COMMAND_BASE_0      = 0x1F0,
        IDE_CONTROLLER_IO_COMMAND_BASE_1      = 0x170,
} IDEControllerIOCommandBase;

typedef enum
{
        IDE_CONTROLLER_IO_CONTROL_BASE_0      = 0x3F0,
        IDE_CONTROLLER_IO_CONTROL_BASE_1      = 0x370
} IDEControllerIOControlBase;

typedef enum
{
        IDE_CONTROLLER_IRQ_0          = 14,
        IDE_CONTROLLER_IRQ_1          = 15
} IDEControllerIRQ;

typedef enum
{
        IDE_CONTROLLER_NUMBER_0       = 0,
        IDE_CONTROLLER_NUMBER_1       = 1,
        IDE_CONTROLLER_COUNT          = 2
} IDEControllerNumber;

enum
{
        IDE_CONTROLLER_COMMAND_TIMEOUT = 20000000
};

typedef struct IDEController
{
        Bool                            isAllocated;
        Bool                            isDetected;
        IDEControllerNumber             number;
        IDEControllerIOCommandBase      ioCommandBase;
        IDEControllerIOControlBase      ioControlBase;
        IDEControllerIRQ                irq;
        InterruptHandler                interruptHandler;
        Timer                           commandTimer;
        Bool                            awaitingAcknowledgement;
        Size                            awaitingSectors;
        Semaphore                       commandAcknowledged;
        IDEError                        commandError;
        Size                            cylinders;
        Size                            heads;
        Size                            sectorsPerTrack;
        Size                            bytesPerSector;
        IDEDrive                        drives [IDE_DRIVE_COUNT];
} IDEController;

INLINE uint8 ide_controller_read_command_register_8 (const IDEController * controller,
                                                     IDECommandRegister registerNumber)
{ return i386_input_8 (controller -> ioCommandBase + registerNumber); }

INLINE uint16 ide_controller_read_command_register_16 (const IDEController * controller,
                                                       IDECommandRegister registerNumber)
{ return i386_input_16 (controller -> ioCommandBase + registerNumber); }

INLINE uint16 ide_controller_read_command_register_32 (const IDEController * controller,
                                                       IDECommandRegister registerNumber)
{ return i386_input_32 (controller -> ioCommandBase + registerNumber); }

INLINE void ide_controller_write_command_register_8 (const IDEController * controller,
                                                     IDECommandRegister registerNumber,
                                                     uint8 data)
{ i386_output_8 (controller -> ioCommandBase + registerNumber, data); }

INLINE void ide_controller_write_command_register_16 (const IDEController * controller,
                                                      IDECommandRegister registerNumber,
                                                      uint16 data)
{ i386_output_32 (controller -> ioCommandBase + registerNumber, data); }

INLINE void ide_controller_write_command_register_32 (const IDEController * controller,
                                                      IDECommandRegister registerNumber,
                                                      uint32 data)
{ i386_output_32 (controller -> ioCommandBase + registerNumber, data); }

INLINE uint8 ide_controller_read_control_register_8 (const IDEController * controller,
                                                     IDEControlRegister registerNumber)
{ return i386_input_8 (controller -> ioControlBase + registerNumber); }

INLINE void ide_controller_write_control_register_8 (const IDEController * controller,
                                                     IDEControlRegister registerNumber,
                                                     uint8 data)
{ i386_output_8 (controller -> ioControlBase + registerNumber, data); }

extern IDEController * ide_controller_detect   (IDEControllerNumber);
extern IDEController * ide_controller_allocate (IDEControllerNumber);
extern void ide_controller_free                (IDEController *);

extern void ide_controller_poll                (const IDEController *, uint8, uint8);
extern void ide_controller_issue_command       (IDEController *, 
                                                IDEDriveNumber, 
                                                IDECommand, 
                                                uint16, 
                                                uint8, 
                                                uint8, 
                                                uint8);
extern void ide_controller_cancel_command     (IDEController *);

#endif /* __IDE_CONTROLLER_H__ */


