#include <i386/manage.h>
#include <scheduler/semaphore.h>
#include <interrupt/pic.h>
#include <interrupt/irq.h>
#include <interrupt/interrupt.h>

Bool
interrupt_handler_create (InterruptHandler * handler,
                          InterruptNumber number,
                          InterruptHandlerFunction * handlerFunction,
                          void * handlerData,
                          InterruptHandlerPriority priority,
                          const char * name)
{
        handler -> handlerFunction = handlerFunction;
        handler -> handlerData     = (handlerData == NULL ? handler : handlerData);
        handler -> priority        = priority;
        handler -> name            = name;
        handler -> number          = IRQ_BASE_INTERRUPTS + number;

        irq_handler_insert ((IRQHandler *) handler);

        return TRUE;
}

void
interrupt_handler_destroy (InterruptHandler * handler)
{
        irq_handler_remove ((IRQHandler *) handler);
}

void
interrupt_enable (InterruptNumber number)
{
        if (number == INTERRUPT_NUMBER_ANY)
          i386_interrupts_enable ();
        else
          pic_unmask_irq (number);
}

void
interrupt_disable (InterruptNumber number)
{
        if (number == INTERRUPT_NUMBER_ANY)
          i386_interrupts_disable ();
        else
          pic_mask_irq (number);
}

static void
interrupt_wait_timeout_handler (Timer * timer, void * interruptDetected)
{
        if (interruptDetected != NULL)
          semaphore_up ((Semaphore *) interruptDetected);
}

static void
interrupt_wait_interrupt_handler (InterruptNumber number, void * interruptDetected)
{
        if (interruptDetected != NULL)
          semaphore_up ((Semaphore *) interruptDetected);
}

Bool
interrupt_wait (InterruptNumber number, TimeOffset * timeoutDelta)
{
        InterruptHandler handler;
        Semaphore interruptDetected;
        Timer interruptTimeout;

        if (number == INTERRUPT_NUMBER_ANY)
        {
           if (timeoutDelta != NULL)
           {
              timer_create (& interruptTimeout,
                            timeoutDelta,
                            interrupt_wait_timeout_handler,
                            NULL,
                            FALSE);

              timer_start (& interruptTimeout);
           }
           
           i386_interrupt_wait ();

           if (timeoutDelta != NULL)
             timer_destroy (& interruptTimeout);

           return TRUE;
        }

        semaphore_create (& interruptDetected, 0);
        
        if (timeoutDelta != NULL)
          timer_create (& interruptTimeout,
                        timeoutDelta,
                        interrupt_wait_timeout_handler,
                        & interruptDetected,
                        FALSE);

        interrupt_handler_create (& handler,
                                  number,
                                  interrupt_wait_interrupt_handler,
                                  & interruptDetected,
                                  ~0,
                                  "interrupt_wait");
        if (timeoutDelta != NULL)
          timer_start (& interruptTimeout);

        semaphore_down (& interruptDetected);
        
        interrupt_handler_destroy (& handler);

        semaphore_destroy (& interruptDetected);

        if (timeoutDelta != NULL) 
        {
           timer_stop (& interruptTimeout, timeoutDelta);
           timer_destroy (& interruptTimeout);

           if (timeoutDelta -> seconds == 0 &&
               timeoutDelta -> nanoSeconds == 0)
             return FALSE;
        }

        return TRUE;
}
        
