#include <i386/manage.h>
#include <scheduler/critical.h>
#include <scheduler/mutex.h>
#include <interrupt/pic.h>
#include <interrupt/irq.h>
#include <interrupt/interrupt.h>
#include <interrupt/exception.h>

static Size  interruptLevel = 0;
static Size  exceptionLevel = 0;
static Mutex irqHandlersMutex;
static List  irqHandlers [IRQ_COUNT];

Bool
__interrupt_enter__ (void)
{
   List * handlerList;

   mutex_create (& irqHandlersMutex);

   for (handlerList = irqHandlers;
        handlerList < & irqHandlers [IRQ_COUNT];
        ++ handlerList)
     list_clear (handlerList);

   irq_initialize ();

   pic_initialize ();

   return TRUE;
}

void
irq_initialize (void)
{
   i386GateDescriptor * idt = (i386GateDescriptor *) I386_IDT_OFFSET,
                      * idtDescriptor;
   IRQStub * irqStub;

   for (idtDescriptor = idt,
           irqStub = irq_stubs;
        idtDescriptor < & idt [IRQ_COUNT];
        ++ idtDescriptor,
           ++ irqStub)
   {
        idtDescriptor -> offset_0   = (uint32) irqStub & 0xFFFF;
        idtDescriptor -> offset_1   = (uint32) irqStub >> 16;
        idtDescriptor -> selector   = I386_SELECTOR_SYSTEM_CODE;
        idtDescriptor -> parameters = 0;
        idtDescriptor -> attributes = I386_DESCRIPTOR_ATTRIBUTE_TYPE_INT_GATE |
                                      I386_DESCRIPTOR_ATTRIBUTE_PRESENT;
   }

   i386_set_IDTR (idt, IRQ_COUNT);
}

Size
interrupt_level (void)
{
   return interruptLevel;
}

void
interrupt_dispatch (IRQRegisters registers)
{
   ListIterator position;
   InterruptHandler * interruptHandler;

   critical_notify_enter ();

   ++ interruptLevel;
   
   position = list_begin (& irqHandlers [registers.irqVector]);
  
   pic_acknowledge_irq (registers.irqVector - IRQ_BASE_INTERRUPTS);

   while (position != list_end (& irqHandlers [registers.irqVector]))
   {
      interruptHandler = (InterruptHandler *) position;

      position = list_next (position);

      (* interruptHandler -> handlerFunction) (registers.irqVector - IRQ_BASE_INTERRUPTS, 
                                               interruptHandler -> handlerData);
   }

   -- interruptLevel;

   critical_notify_exit ();
}

Size
exception_level (void)
{
   return exceptionLevel;
}

void
exception_dispatch (IRQRegisters registers)
{
   ListIterator position;
   ExceptionHandler * exceptionHandler;

   critical_notify_enter ();

   ++ exceptionLevel;

   position = list_begin (& irqHandlers [registers.irqVector]);

   while (position != list_end (& irqHandlers [registers.irqVector]))
   {
      exceptionHandler = (ExceptionHandler *) position;

      position = list_next (position);

      (* exceptionHandler -> handlerFunction) (registers.irqVector - IRQ_BASE_EXCEPTIONS,
                                               registers.errorCode,
                                               exceptionHandler -> handlerData);
   }
  
   -- exceptionLevel;

   critical_notify_exit ();
}

void
irq_handler_insert (IRQHandler * irqHandler)
{
   ListIterator position;
   
   mutex_lock (& irqHandlersMutex);
   
   irqHandler -> handlerList.key = irqHandler -> priority;

   position = list_find_key (& irqHandlers [irqHandler -> number], irqHandler -> priority);

   list_insert (position, irqHandler);

   if (position == list_end (& irqHandlers [irqHandler -> number]));
     interrupt_enable (irqHandler -> number - IRQ_BASE_INTERRUPTS);

   mutex_unlock (& irqHandlersMutex);
}

void
irq_handler_remove (IRQHandler * irqHandler)
{
   mutex_lock (& irqHandlersMutex);

   list_remove ((ListIterator) & irqHandler -> handlerList);

   if (list_empty (& irqHandlers [irqHandler -> number]) &&
       irqHandler -> number >= IRQ_BASE_INTERRUPTS &&
       irqHandler -> number <= IRQ_BASE_SOFTWARE)
     interrupt_disable (irqHandler -> number - IRQ_BASE_INTERRUPTS);
   
   mutex_unlock (& irqHandlersMutex);
}

