#include <console/console.h>
#include <scheduler/mutex.h>
#include <kbc/kbc.h>

static Mutex kbcMutex;

Bool
__kbc_enter__ (void)
{
    mutex_create (& kbcMutex);

    if (kbc_reset () == FALSE)
      return FALSE;

    return TRUE;
}

void
__kbc_exit__ (void)
{
    mutex_destroy (& kbcMutex);
}

void
kbc_send_command (uint8 command)
{
    kbc_wait_for_write ();
    kbc_clear_data ();
    kbc_write_command (command);
}

uint8
kbc_receive_data (void)
{
    kbc_wait_for_read ();

    return kbc_read_data ();
}

Bool
kbc_acknowledge_data (uint8 data)
{
    uint8 acknowledgement;
    
    do
    {
       kbc_wait_for_write ();
       kbc_clear_data ();
       kbc_write_data (data);

       acknowledgement = kbc_receive_data ();
       if (acknowledgement == KBC_DATA_COMMAND_ACKNOWLEDGE)
         return TRUE;
    } while (acknowledgement == KBC_DATA_RESEND_REQUEST);
    
    return FALSE;
} 

Bool
kbc_reset (void)
{
    console_print (NULL, "KBC: Resetting keyboard controller...\n");

    kbc_send_command (KBC_COMMAND_ACTIVATE_KEYBOARD);

    if (kbc_acknowledge_data (KBC_COMMAND_DISABLE_SCANNING) == FALSE)
      return FALSE;
    
    console_print (NULL, "KBC: Starting self test...\n");

    kbc_send_command (KBC_COMMAND_SELF_TEST);

    if (kbc_receive_data () != KBC_DATA_SELF_TEST_SUCCEEDED)
    {
       console_print (NULL, "KBC: Self test failed!\n");
       
       return FALSE;
    }

    console_print (NULL, "KBC: Testing keyboard interface...\n");

    kbc_send_command (KBC_COMMAND_TEST_KEYBOARD_INTERFACE);
    if (kbc_receive_data () != 0)
    {
       console_print (NULL, "KBC: Keyboard interface test failed!\n");
       
       return FALSE;
    }

    kbc_send_command (KBC_COMMAND_ACTIVATE_KEYBOARD);

    console_print (NULL, "KBC: Starting basic assurance test...\n");

    kbc_acknowledge_data (KBC_COMMAND_INTERNAL_DIAGNOSTICS);
    if (kbc_receive_data () != KBC_DATA_BASIC_ASSURANCE_TEST_SUCCEEDED)
    {
       console_print (NULL, "KBC: Basic assurance test failed!\n");
       
       return FALSE;
    }

    kbc_send_command (KBC_COMMAND_DEACTIVATE_KEYBOARD);
    kbc_send_command (KBC_COMMAND_WRITE_COMMAND_BYTE);
    kbc_wait_for_write ();
    kbc_write_data (KBC_COMMAND_BYTE_SYSTEM_RESERVED_FLAG | KBC_COMMAND_BYTE_AUX_DISABLED);
    
    console_print (NULL, "KBC: Keyboard controller reset.\n");

    return TRUE;
}

void
kbc_lock (void)
{
    mutex_lock (& kbcMutex);
}

void
kbc_unlock (void)
{
    mutex_unlock (& kbcMutex);
}

