#include <utility/string.h>
#include <memory/memory.h>
#include <linker/linker.h>
#include <linker/constructor.h>
#include <linker/destructor.h>
#include <linker/component.h>

Bool
component_create (Component ** result, const char * name)
{
        Component * component = (Component *) memory_allocate (sizeof (Component));

        if (component == NULL)
          return FALSE;

        component -> linker = linker_current ();
        component -> state  = COMPONENT_STATE_CONSTRUCTING;
        component -> name   = name;

        list_clear (& component -> sections);
        list_clear (& component -> interfaces);
        list_clear (& component -> constructors);
        list_clear (& component -> destructors);
        list_clear (& component -> imports);

        hash_table_create (& component -> exports, 1);

        component -> referenceCount = 0;

        list_insert_back (& component -> linker -> components, component);
       
        if (result != NULL)
          * result = component;
        
        return TRUE;
}

void
component_destroy (Component * component)
{
        if (component -> referenceCount > 0)
        {
           component -> state = COMPONENT_STATE_DESTRUCTING;

           return;
        }
 
        if (component -> state != COMPONENT_STATE_CONSTRUCTING)
          component_call_destructors (component);
        else
          component -> state = COMPONENT_STATE_DESTRUCTING;

        list_remove (& component -> componentList);

        while (list_empty (& component -> constructors) == FALSE)
          constructor_destroy ((Constructor *) list_front (& component -> constructors));

        while (list_empty (& component -> destructors) == FALSE)
          destructor_destroy ((Destructor *) list_front (& component -> destructors));

        while (list_empty (& component -> sections) == FALSE)
          section_destroy ((Section *) list_front (& component -> sections));

        while (list_empty (& component -> imports) == FALSE)
          import_destroy ((Import *) list_front (& component -> imports));

        while (list_empty (& component -> interfaces) == FALSE)
          interface_destroy ((Interface *) list_front (& component -> interfaces));

        hash_table_destroy (& component -> exports);

        memory_free (component);
}
         
Export *
component_find_export (Component * component, const char * name)
{
        void ** values = hash_table_find (& component -> exports, name);

        if (values != NULL)
          return (Export *) * values;

        return NULL;
}

Import *
component_find_import (Component * component, const char * interface, const char * name)
{
        ListIterator position;

        for (position = list_begin (& component -> imports);
             position != list_end (& component -> imports);
             position = list_next (position))
        {
            Import * import = (Import *) position;

            if (interface != NULL &&
                (import -> interface == NULL ||
                 string_equal (import -> interface, interface) == FALSE))
              continue;
            
            if (string_equal (import -> name, name) == TRUE)
              return import;
        }

        return NULL;
}
                
Interface *
component_find_interface (Component * component, const char * name)
{
        ListIterator position;

        for (position = list_begin (& component -> interfaces);
             position != list_end (& component -> interfaces);
             position = list_next (position))
        {
            Interface * interface = (Interface *) position;

            if (string_equal (interface -> name, name) == TRUE)
              return interface;
        }

        return NULL;
}

Section *
component_find_section (Component * component, const char * name)
{
        ListIterator position;

        for (position = list_begin (& component -> sections);
             position != list_end (& component -> sections);
             position = list_next (position))
        {
            Section * section = (Section *) position;

            if (string_equal (section -> name, name) == TRUE)
              return section;
        }

        return NULL;
}

void
component_relocate_sections (Component * component)
{
        ListIterator position;

        for (position = list_begin (& component -> sections);
             position != list_end (& component -> sections);
             position = list_next (position))
        {
            Section * section = (Section *) position;

            section_relocate (section);
        }
}

Bool
component_resolve_imports (Component * component)
{
        ListIterator position;

        for (position = list_begin (& component -> imports);
             position != list_end (& component -> imports);
             position = list_next (position))
        {
            Import * import = (Import *) position;
            Export * export;

            if (import -> interface == NULL)
              export = component_find_export (component, import -> name);
            else
              export = linker_find_export (import -> interface, import -> name);
           
            if (export == NULL)
              return FALSE;

            if (import -> type != IMPORT_TYPE_UNKNOWN &&
                import -> type != (ImportType) export -> type)
              return FALSE;

            import -> export = export;
        }
        
        return TRUE;
}

Bool
component_call_constructors (Component * component)
{
        ListIterator position;

        for (position = list_begin (& component -> constructors);
             position != list_end (& component -> constructors);
             position = list_next (position))
        {
            Constructor * constructor = (Constructor *) position;
            Export * export           = component_find_export (component, constructor -> name);
            ConstructorFunction * constructorFunction;

            if (export == NULL ||
                export -> interface != NULL ||
                export -> type != EXPORT_TYPE_FUNCTION)
              return FALSE;

            constructorFunction = (ConstructorFunction *) (export -> section -> data + export -> offset);
            
            if ((* constructorFunction) () == FALSE)
              return FALSE;
        }
       
        component -> state = COMPONENT_STATE_READY;

        return TRUE;
}

void
component_call_destructors (Component * component)
{
        ListIterator position;

        component -> state = COMPONENT_STATE_DESTRUCTING;

        for (position = list_begin (& component -> constructors);
             position != list_end (& component -> constructors);
             position = list_next (position))
        {
            Destructor * destructor = (Destructor *) position;
            Export * export         = component_find_export (component, destructor -> name);
            DestructorFunction * destructorFunction;

            if (export == NULL ||
                export -> interface != NULL ||
                export -> type != EXPORT_TYPE_FUNCTION)
              continue;

            destructorFunction = (DestructorFunction *) (export -> section -> data + export -> offset);
            
            (* destructorFunction) ();            
        }
}

