#include <utility/memory.h>
#include <utility/string.h>
#include <linker/interface.h>
#include <linker/constructor.h>
#include <linker/destructor.h>
#include <linker/pcom.h>

Bool
pcom_create (PCom * pcom, uint8 * data, Size dataSize)
{
        PComHeader * header = (PComHeader *) data;
        PComSection * stringsSection;

        if (dataSize < sizeof (PComHeader) ||
            pcom_validate (header) == FALSE ||
            hash_table_create (& pcom -> symbols, 4) == FALSE)
          return FALSE;

        stringsSection = (PComSection *) & header [1] + header -> stringsSection;
        
        pcom -> header   = header;
        pcom -> strings  = (const char *) & data [stringsSection -> offset];
        pcom -> data     = data;
        pcom -> dataSize = dataSize;
        pcom -> name     = & pcom -> strings [header -> name];
       
        return TRUE;
}

void
pcom_destroy (PCom * pcom)
{
        hash_table_destroy (& pcom -> symbols);
}

Bool
pcom_find_symbol (PCom * pcom, 
                  const char * symbolName,
                  PComSectionType * sectionType,
                  const char ** sectionName,
                  const char ** symbolAlias,
                  void ** data)
{
        void ** values = hash_table_find (& pcom -> symbols, symbolName);

        if (values == NULL)
          return FALSE;

        if (sectionType != NULL)
          * sectionType = (PComSectionType) values [0];
        
        if (sectionName != NULL)
          * sectionName = (const char *) values [1];

        if (symbolAlias != NULL)
          * symbolAlias = (const char *) values [2];

        if (data != NULL)
          * data = values [3];
        
        return TRUE;
}

Bool
pcom_validate (PComHeader * header)
{
        if (memory_compare (header -> signature, PCOM_SIGNATURE, PCOM_SIGNATURE_LENGTH) != COMPARE_EQUAL)
          return FALSE;

        if (header -> version != PCOM_VERSION)
          return FALSE;
        
        if (header -> name == PCOM_STRING_NULL)
          return FALSE;

        if (header -> stringsSection == PCOM_SECTION_NUMBER_NULL)
          return FALSE;

        return TRUE;
}

PComSection *
pcom_find_section (PCom * pcom,
                   PComSectionType type,
                   const char * name)
{
        PComSection * sections,
                    * currentSection;

        sections = (PComSection *) & pcom -> header [1];

        for (currentSection = sections;
             currentSection < & sections [pcom -> header -> sectionCount];
             ++ currentSection)
        {
            if (name != NULL &&
                currentSection -> name != PCOM_STRING_NULL &&
                string_equal (& pcom -> strings [currentSection -> name], name) == TRUE)
              return currentSection;

            if (currentSection -> type == type)
              return currentSection;
        }

        return NULL;
}

Bool
pcom_read_symbols (PCom * pcom, PComSection * section, void * data)
{
        const char * sectionName = NULL;
        PComSymbol * symbols,
                   * currentSymbol;

        if (section -> type == PCOM_SECTION_TYPE_STRINGS ||
            section -> symbolCount < 0)
          return TRUE;

        if (section -> name != PCOM_STRING_NULL)
          sectionName = & pcom -> strings [section -> name];

        symbols = (PComSymbol *) & pcom -> data [section -> offset];

        for (currentSymbol = symbols;
             currentSymbol < & symbols [section -> symbolCount];
             ++ currentSymbol)
        {
            const char * symbolName,
                       * symbolAlias;
            void ** values;

            if (currentSymbol -> name == PCOM_STRING_NULL)
              return FALSE;

            symbolName = & pcom -> strings [currentSymbol -> name];
                
            if (currentSymbol -> alias == PCOM_STRING_NULL)
              symbolAlias = symbolName;
            else
              symbolAlias = & pcom -> strings [currentSymbol -> alias];
                
            values = hash_table_insert (& pcom -> symbols, symbolName);

            if (values == NULL)
              return FALSE;

            values [0] = (void *) section -> type;
            values [1] = (void *) sectionName;
            values [2] = (void *) symbolAlias;
            values [3] = data;
        }

        return TRUE;
}

Bool
pcom_add_constructors (PCom * pcom, Component * component)
{
        PComSection * section;
        PComSymbol * symbols,
                   * currentSymbol;

        if (pcom -> header -> constructorsSection == PCOM_SECTION_NUMBER_NULL)
          return TRUE;

        section = (PComSection *) & pcom -> header [1] + pcom -> header -> constructorsSection;
        symbols = (PComSymbol *) & pcom -> data [section -> offset];

        for (currentSymbol = symbols;
             currentSymbol < & symbols [section -> symbolCount];
             ++ currentSymbol)
        {
            const char * name,
                       * alias;
            void ** values;

            if (currentSymbol -> name == PCOM_STRING_NULL)
              return FALSE;

            name = & pcom -> strings [currentSymbol -> alias];

            if (currentSymbol -> alias != PCOM_STRING_NULL)
              alias = & pcom -> strings [currentSymbol -> alias];
            else
              alias = name;

            values = hash_table_insert (& pcom -> symbols, name);

            if (values == NULL ||
                constructor_create ((Constructor **) & values [3], component, name) == FALSE)
              return FALSE;

            values [0] = (void *) PCOM_SECTION_TYPE_CONSTRUCTORS;
            values [1] = NULL;
            values [2] = (void *) alias;
        }

        return TRUE;
}

Bool
pcom_add_destructors (PCom * pcom, Component * component)
{
        PComSection * section;
        PComSymbol * symbols,
                   * currentSymbol;

        if (pcom -> header -> destructorsSection == PCOM_SECTION_NUMBER_NULL)
          return TRUE;

        section = (PComSection *) & pcom -> header [1] + pcom -> header -> destructorsSection;
        symbols = (PComSymbol *) & pcom -> data [section -> offset];

        for (currentSymbol = symbols;
             currentSymbol < & symbols [section -> symbolCount];
             ++ currentSymbol)
        {
            const char * name,
                       * alias;
            void ** values;
            
            if (currentSymbol -> name == PCOM_STRING_NULL)
              return FALSE;

            name = & pcom -> strings [currentSymbol -> alias];

            if (currentSymbol -> alias != PCOM_STRING_NULL)
              alias = & pcom -> strings [currentSymbol -> alias];
            else
              alias = name;

            values = hash_table_insert (& pcom -> symbols, name);
            
            if (values == NULL ||
                destructor_create ((Destructor **) & values [3], component, name) == FALSE)
              return FALSE;
            
            values [0] = (void *) PCOM_SECTION_TYPE_DESTRUCTORS;
            values [1] = NULL;
            values [2] = (void *) alias;
        }

        return TRUE;
}

Bool
pcom_add_exports (PCom * pcom, Component * component)
{
        PComSection * sections,
                    * currentSection;

        sections = (PComSection *) & pcom -> header [1];

        for (currentSection = sections;
             currentSection < & sections [pcom -> header -> sectionCount];
             ++ currentSection)
        {
            Interface * interface;

            if (currentSection -> type != PCOM_SECTION_TYPE_EXPORTS)
              continue;
        
            if (currentSection -> name == PCOM_STRING_NULL ||
                interface_create (& interface,
                                  component,
                                  & pcom -> strings [currentSection -> name]) == FALSE)
              return FALSE;
           
            if (pcom_read_symbols (pcom, currentSection, interface) == FALSE)
              return FALSE;
        }

        return TRUE;
}

Bool
pcom_add_imports (PCom * pcom, Component * component)
{
        PComSection * sections,
                    * currentSection;

        sections = (PComSection *) & pcom -> header [1];

        for (currentSection = sections;
             currentSection < & sections [pcom -> header -> sectionCount];
             ++ currentSection)
        {
            if (currentSection -> type != PCOM_SECTION_TYPE_IMPORTS)
              continue;

            if (pcom_read_symbols (pcom, currentSection, NULL) == FALSE)
              return FALSE;
        }

        return TRUE;
}

