#include <scheduler/mutex.h>
#include <page/allocator.h>

static Mutex            allocatorMutex;
static PageAllocator * 	allocatorCache = NULL;
static List	 	allocatorList;

Bool
__page_enter__ (void)
{
   mutex_create (& allocatorMutex);

   list_clear (& allocatorList);

   return TRUE;
}

void
__page_exit__ (void)
{
   while (list_empty (& allocatorList) == FALSE)
      page_allocator_destroy ((PageAllocator *) list_front (& allocatorList));
}

void
page_allocator_insert (PageAllocator * allocator)
{
   ListIterator position;
   
   mutex_lock (& allocatorMutex);
   
   position = list_find_key (& allocatorList, allocator -> allocatorList.key);

   list_insert (position, allocator);

   mutex_unlock (& allocatorMutex);
}

void
page_allocator_remove (PageAllocator * allocator)
{
   mutex_lock (& allocatorMutex);

   list_remove ((ListIterator) & allocator -> allocatorList);
   
   if (allocator == allocatorCache)
     allocatorCache = NULL;

   mutex_unlock (& allocatorMutex);
}

PageAllocator *
page_allocator_find_range (PageAllocator * allocator, PageAddress rangeBegin, PageAddress rangeEnd)
{
    ListIterator current;

    if (allocatorCache != NULL &&
	allocatorCache != allocator &&
	((rangeBegin >= allocatorCache -> rangeBegin &&
	  rangeBegin < allocatorCache -> rangeEnd) ||
	 (rangeEnd >= allocatorCache -> rangeBegin &&
	  rangeEnd < allocatorCache -> rangeEnd)))
      return allocatorCache;

    for (current  = (allocator == NULL) ? list_begin (& allocatorList) : list_next ((ListIterator) & allocator -> allocatorList);
	 current != list_end (& allocatorList);
	 current  = list_next (current))
      if ((rangeBegin >= ((PageAllocator *) current) -> rangeBegin &&
	   rangeBegin < ((PageAllocator *) current) -> rangeEnd) ||
	  (rangeEnd >= ((PageAllocator *) current) -> rangeBegin &&
	   rangeEnd < ((PageAllocator *) current) -> rangeEnd))
      {
	  allocatorCache = (PageAllocator *) current;

	  return (PageAllocator *) current;
      }
    
    return NULL;
}

PageAllocator *
page_allocator_find_attributes (PageAllocator * allocator, PageAllocatorAttributes attributes)
{
    ListIterator current;

    if (allocatorCache != NULL &&
	allocatorCache != allocator &&
        (allocatorCache -> attributes & attributes) == attributes)
      return allocatorCache;

    for (current  = (allocator == NULL) ? list_begin (& allocatorList) : list_next ((ListIterator) & allocator -> allocatorList);
         current != list_end (& allocatorList);
	 current  = list_next (current))
      if ((((PageAllocator *) current) -> attributes & attributes) == attributes)
      {
          allocatorCache = (PageAllocator *) current;
	  
          return (PageAllocator *) current;
      }

    return NULL;
}

PageAllocator *
page_allocator_find_address (PageAllocator * allocator, PageAddress address)
{
    ListIterator current;

    if (allocatorCache != NULL &&
        allocatorCache != allocator &&		
        address >= allocatorCache -> rangeBegin &&
	address <  allocatorCache -> rangeEnd)
      return allocatorCache;

    for (current  = (allocator == NULL) ? list_begin (& allocatorList) : list_next ((ListIterator) & allocator -> allocatorList);
         current != list_end (& allocatorList);
	 current  = list_next (current))
      if (address >= ((PageAllocator *) current) -> rangeBegin &&
          address <  ((PageAllocator *) current) -> rangeEnd)
      {
          allocatorCache = (PageAllocator *) current;
    
          return (PageAllocator *) current;
      }

    return NULL;
}

PageAddress
page_allocate_within (PageAlignment alignment,
		      PageAddress rangeBegin,
		      PageAddress rangeEnd,
		      void * owner)
{
     PageAllocator * allocator;
     Page * page = NULL;

     mutex_lock (& allocatorMutex);

     for (allocator = page_allocator_find_range (NULL, rangeBegin, rangeEnd);
	  allocator != NULL;
	  allocator = page_allocator_find_range (allocator, rangeBegin, rangeEnd))
     {
	  page = page_allocator_allocate_within (allocator, alignment, rangeBegin, rangeEnd, owner);
          if (page != NULL)
            break;

	  if (page_allocator_fill (allocator, alignment, ~0) <= 0)
            continue;

	  page = page_allocator_allocate_within (allocator, alignment, rangeBegin, rangeEnd, owner);
	  if (page != NULL)
	    break;
     }

     mutex_unlock (& allocatorMutex);
     
     if (page != NULL)
       return page -> address;

     return 0;
}

PageAddress
page_allocate (PageAlignment alignment, 
	       PageAllocatorAttributes attributes,
	       void * owner)
{
     PageAllocator * allocator;
     Page * page = NULL;

     mutex_lock (& allocatorMutex);

     for (allocator = page_allocator_find_attributes (NULL, attributes);
	  allocator != NULL;
	  allocator = page_allocator_find_attributes (allocator, attributes))
     {
         page = page_allocator_allocate (allocator, alignment, owner);
	 if (page != NULL)
           break;

	 if (page_allocator_fill (allocator, alignment, ~0) <= 0)
           continue;
        
	 page = page_allocator_allocate (allocator, alignment, owner);
	 if (page != NULL)
           break;
     }

     mutex_unlock (& allocatorMutex);
     
     if (page != NULL)
       return page -> address;

     return 0;
}

void
page_free (PageAddress address)
{
     PageAllocator * allocator;
     
     mutex_lock (& allocatorMutex);
    
     allocator = page_allocator_find_address (NULL, address);

     if (allocator != NULL)
       page_allocator_free (allocator, page_allocator_find (allocator, address));

     mutex_unlock (& allocatorMutex);
}

Size
page_get_size (PageAddress address)
{
     PageAllocator * allocator;
     Page * page = NULL;

     mutex_lock (& allocatorMutex);

     allocator = page_allocator_find_address (NULL, address);
     
     if (allocator != NULL)
       page = page_allocator_find (allocator, address);

     mutex_unlock (& allocatorMutex);

     if (page != NULL &&
	 page -> owner != NULL)
       return (1 << page -> alignment);

     return 0;
}

void *
page_get_owner (PageAddress address)
{
     PageAllocator * allocator;
     Page * page = NULL;

     mutex_lock (& allocatorMutex);

     allocator = page_allocator_find_address (NULL, address);

     if (allocator != NULL)
       page = page_allocator_find (allocator, address);

     mutex_unlock (& allocatorMutex);

     if (page != NULL)
       return page -> owner;

     return NULL;
}

