#include <console/console.h>
#include <memory/memory.h>
#include <pci/device.h>
#include <pci/bios.h>

static List pciDevices;

Bool
__pci_enter__ (void)
{
    PCIBiosVersion majorVersion, minorVersion;
    PCIBusNumber lastBus;

    list_clear (& pciDevices);

    if (pci_bios_initialize () == FALSE)
    {
       console_print (NULL, "PCI BIOS: Initialization failed!\n");

       return FALSE;
    }

    if (pci_bios_probe (& majorVersion, & minorVersion, & lastBus) != PCI_BIOS_ERROR_SUCCESSFUL)
    {
       console_print (NULL, "PCI BIOS: Probing failed!\n");

       return FALSE;
    }

    console_print_format (NULL,
                          "PCI BIOS: Probing...\n"
                          "PCI BIOS:   Major version: %x\n"
                          "PCI BIOS:   Minor version: %x\n"
                          "PCI BIOS:   Last bus: %d\n",
                          majorVersion,
                          minorVersion,
                          lastBus);

    pci_device_list (& pciDevices);

    return TRUE;
}

void
__pci_exit__ (void)
{
    while (list_empty (& pciDevices) == FALSE)
       memory_free (list_remove_front (& pciDevices));
}

void
pci_device_list (List * list)
{
   PCIBiosVersion majorVersion, minorVersion;
   PCIBusNumber busNumber, lastBus;
   PCIDeviceNumber deviceNumber;
   PCIDevice * currentDevice;

   list_clear (list);

   if (pci_bios_probe (& majorVersion, & minorVersion, & lastBus) != PCI_BIOS_ERROR_SUCCESSFUL)
     return;

   currentDevice = (PCIDevice *) memory_allocate (sizeof (PCIDevice));

   for (busNumber = 0;
        busNumber <= lastBus;
	++ busNumber)
   {
       for (deviceNumber = 0;
            deviceNumber <= PCI_DEVICE_NUMBER_MAXIMUM;
	    ++ deviceNumber)
       {
           if (pci_bios_read_config_16 (busNumber, 
				        deviceNumber,
					PCI_CONFIGURATION_VENDOR_ID,
					& currentDevice -> vendorID) != PCI_BIOS_ERROR_SUCCESSFUL ||
	       currentDevice -> vendorID == 0 ||
	       currentDevice -> vendorID == PCI_VENDOR_ID_ANY)
	    continue;

	  pci_bios_read_config_16 (busNumber,
			           deviceNumber,
				   PCI_CONFIGURATION_DEVICE_ID,
				   & currentDevice -> deviceID);
	  pci_bios_read_config_16 (busNumber,
			           deviceNumber,
				   PCI_CONFIGURATION_SUBSYSTEM_VENDOR_ID,
				   & currentDevice -> subsystemVendorID);
	  pci_bios_read_config_16 (busNumber,
			           deviceNumber,
				   PCI_CONFIGURATION_SUBSYSTEM_DEVICE_ID,
				   & currentDevice -> subsystemDeviceID);
	  pci_bios_read_config_32 (busNumber,
			           deviceNumber,
				   PCI_CONFIGURATION_CLASS_CODE,
				   & currentDevice -> classCode);

          currentDevice -> classCode >>= 8;

          console_print_format (NULL, 
                                "PCI: Bus %u, Device %u...\n"
                                "PCI:   Vendor ID: 0x%x\n"
                                "PCI:   Device ID: 0x%x\n"
                                "PCI:   Subsystem Vendor ID: 0x%x\n"
                                "PCI:   Subsystem Device ID: 0x%x\n"
                                "PCI:   Class Code: 0x%x\n",
                                busNumber,
                                deviceNumber,
                                currentDevice -> vendorID,
                                currentDevice -> deviceID,
                                currentDevice -> subsystemVendorID,
                                currentDevice -> subsystemDeviceID,
                                currentDevice -> classCode);

	  currentDevice -> isAllocated = FALSE;

	  list_insert_front (list, currentDevice);

	  currentDevice = (PCIDevice *) memory_allocate (sizeof (PCIDevice));
       }
   }

   memory_free (currentDevice);
}

const PCIDevice *
pci_device_find (const PCIDevice * device,
                 PCIVendorID vendorID,
                 PCIDeviceID deviceID)
{
   if (device == NULL)
     device = (const PCIDevice *) list_begin (& pciDevices);
   else
     device = (const PCIDevice *) list_next (& device -> deviceList);
      
   while (& device -> deviceList != list_end (& pciDevices))
   {
      if ((vendorID == PCI_VENDOR_ID_ANY ||
           vendorID == device -> vendorID) &&
          (deviceID == PCI_DEVICE_ID_ANY ||
           deviceID == device -> deviceID))
        return device;

      device = (const PCIDevice *) list_next (& device -> deviceList);
   }

   return NULL;
}

const PCIDevice *
pci_device_find_subsystem (const PCIDevice * device, 
		           PCIVendorID vendorID, 
		           PCIDeviceID deviceID)
{
   if (device == NULL)
     device = (const PCIDevice *) list_begin (& pciDevices);
   else
     device = (const PCIDevice *) list_next (& device -> deviceList);
                 
   while (& device -> deviceList != list_end (& pciDevices))
   {
      if ((vendorID == PCI_VENDOR_ID_ANY ||
           vendorID == device -> subsystemVendorID) &&
	  (deviceID == PCI_DEVICE_ID_ANY ||
	   deviceID == device -> subsystemDeviceID))
	return device;
   
      device = (const PCIDevice *) list_next (& device -> deviceList);
   }

   return NULL;
}

const PCIDevice *
pci_find_class (const PCIDevice * device, PCIClassCode classCode)
{
   if (device == NULL)
     device = (const PCIDevice *) list_begin (& pciDevices);
   else
     device = (const PCIDevice *) list_next (& device -> deviceList);
                 
   while (& device -> deviceList != list_end (& pciDevices))
   {
      if (classCode == device -> classCode)
	return device;

      device = (const PCIDevice *) list_next (& device -> deviceList);
   }

   return NULL;
}

Bool
pci_device_read_config_8 (const PCIDevice * device,
                          PCIRegisterNumber registerNumber,
			  uint8 * value)
{
   if (pci_bios_read_config_8 (device -> busNumber,
                               device -> deviceNumber,
                               registerNumber,
                               value) != PCI_BIOS_ERROR_SUCCESSFUL)
     return FALSE;

   return TRUE;
}

Bool
pci_device_read_config_16 (const PCIDevice * device,
                           PCIRegisterNumber registerNumber,
                           uint16 * value)
{
   if (pci_bios_read_config_16 (device -> busNumber,
                                device -> deviceNumber,
                                registerNumber,
                                value) != PCI_BIOS_ERROR_SUCCESSFUL)
     return FALSE;

   return TRUE;
}

Bool
pci_device_read_config_32 (const PCIDevice * device,
                           PCIRegisterNumber registerNumber,
                           uint32 * value)
{
   if (pci_bios_read_config_32 (device -> busNumber,
                                device -> deviceNumber,
                                registerNumber,
                                value) != PCI_BIOS_ERROR_SUCCESSFUL)
     return FALSE;

   return TRUE;
}

Bool
pci_device_write_config_8 (const PCIDevice * device,
                           PCIRegisterNumber registerNumber,
                           uint8 value)
{
   if (pci_bios_write_config_8 (device -> busNumber,
                                device -> deviceNumber,
                                registerNumber,
                                value) != PCI_BIOS_ERROR_SUCCESSFUL)
     return FALSE;

   return TRUE;
}

Bool
pci_device_write_config_16 (const PCIDevice * device,
                            PCIRegisterNumber registerNumber,
                            uint16 value)
{
   if (pci_bios_write_config_16 (device -> busNumber,
                                 device -> deviceNumber,
                                 registerNumber,
                                 value) != PCI_BIOS_ERROR_SUCCESSFUL)
     return FALSE;

   return TRUE;
}

Bool
pci_device_write_config_32 (const PCIDevice * device,
                            PCIRegisterNumber registerNumber,
                            uint32 value)
{
   if (pci_bios_write_config_32 (device -> busNumber,
                                 device -> deviceNumber,
                                 registerNumber,
                                 value) != PCI_BIOS_ERROR_SUCCESSFUL)
     return FALSE;

   return TRUE;
}

PCIDevice *
pci_device_allocate (const PCIDevice * device)
{
   PCIDevice * result;
   
   if (device -> isAllocated == TRUE)
     return NULL;

   result = (PCIDevice *) device;

   result -> isAllocated = TRUE;

   return result;
}

void
pci_device_free (PCIDevice * device)
{
   device -> isAllocated = FALSE;
}

