#include <scheduler/critical.h>
#include <scheduler/thread.h>
#include <scheduler/condition.h>

Bool
condition_create (Condition * condition)
{
   spinlock_create (& condition -> lock);

   list_clear (& condition -> waitingThreads);
 
   return TRUE;
}

void
condition_destroy (Condition * condition)
{
   condition_broadcast (condition);

   spinlock_destroy (& condition -> lock);
}

Bool
condition_wait (Condition * condition, Mutex * mutex)
{
   spinlock_lock (& condition -> lock);

   if (mutex_unlock (mutex) >= ~0)
   {
      spinlock_unlock (& condition -> lock);

      return FALSE;
   }

   thread_block (NULL, THREAD_STATE_WAITING_ON_CONDITION, & condition -> lock, & condition -> waitingThreads);

   if (mutex_lock (mutex) <= 0)
   {
      spinlock_unlock (& condition -> lock);

      return FALSE;
   }

   spinlock_unlock (& condition -> lock);

   return TRUE;
}

typedef struct
{
   Thread *    thread;
   Condition * condition;
} TimedWaitContext;

static void
timedWaitTimeoutHandler (Timer * timer, void * data)
{
   TimedWaitContext * context = (TimedWaitContext *) data;

   spinlock_lock (& context -> condition -> lock);

   list_remove (& context -> thread -> threadList);

   thread_start (context -> thread);
  
   spinlock_unlock (& context -> condition -> lock);
}

Bool
condition_timed_wait (Condition * condition, Mutex * mutex, TimeOffset * timeout)
{
   Timer timer;
   TimedWaitContext context;

   context.thread    = thread_current ();
   context.condition = condition;

   spinlock_lock (& condition -> lock);

   if (mutex_unlock (mutex) >= ~0)
   {
      spinlock_unlock (& condition -> lock);

      return FALSE;
   }

   timer_create (& timer, timeout, timedWaitTimeoutHandler, & context, FALSE);

   critical_enter ();

   timer_start (& timer);

   thread_block (NULL, THREAD_STATE_WAITING_ON_CONDITION, & condition -> lock, & condition -> waitingThreads);

   timer_stop (& timer, timeout);

   critical_exit ();

   if (mutex_lock (mutex) <= 0)
   {
      spinlock_unlock (& condition -> lock);

      timer_destroy (& timer);

      return FALSE;
   }
   
   spinlock_unlock (& condition -> lock);

   timer_destroy (& timer);

   return TRUE;
}

Bool
condition_signal (Condition * condition)
{
   spinlock_lock (& condition -> lock);

   if (list_empty (& condition -> waitingThreads) == FALSE)
   {
      Thread * waiting = (Thread *) list_remove_front (& condition -> waitingThreads);

      thread_start (waiting);

      spinlock_unlock (& condition -> lock);

      return TRUE;
   }

   spinlock_unlock (& condition -> lock);

   return FALSE;
}

Size
condition_broadcast (Condition * condition)
{
   Size result = 0;

   spinlock_lock (& condition -> lock);

   while (list_empty (& condition -> waitingThreads) == FALSE)
   {
      Thread * waiting = (Thread *) list_remove_front (& condition -> waitingThreads);

      thread_start (waiting);

      ++ result;
   }

   spinlock_unlock (& condition -> lock);

   return result;
}

