#include <interrupt/interrupt.h>
#include <interrupt/exception.h>
#include <timer/timer.h>
#include <memory/memory.h>
#include <scheduler/registers.h>
#include <scheduler/critical.h>
#include <scheduler/thread.h>
#include <scheduler/scheduler.h>

static Scheduler   schedulerDefault;
static Scheduler * schedulerCurrent = NULL;

Thread *
thread_current (void)
{
   if (schedulerCurrent != NULL)
     return schedulerCurrent -> threadCurrent;

   return NULL;
}

Scheduler *
scheduler_current (void)
{
   return schedulerCurrent;
}

static void
scheduler_idle_handler (void * data)
{
   for (;;)
     interrupt_wait (INTERRUPT_NUMBER_ANY, NULL);
}

static void
scheduler_timer_handler (Timer * timer, void * data)
{
   Scheduler * scheduler = (Scheduler *) data;

   if (scheduler -> threadCurrent == NULL ||
       scheduler -> threadCurrent -> priorityCurrent <= THREAD_PRIORITY_IDLE ||
       scheduler -> threadCurrent -> priorityCurrent >= THREAD_PRIORITY_EXCLUSIVE)
     return;
   
   -- scheduler -> threadCurrent -> quantumCurrent;
   if (scheduler -> threadCurrent -> quantumCurrent <= 0)
   {
      if (scheduler -> threadCurrent -> priorityCurrent > scheduler -> threadCurrent -> priorityFloor)
        thread_set_priority (scheduler -> threadCurrent,
			     scheduler -> threadCurrent -> priorityCurrent - 1);

      thread_yield (NULL);

      /*thread_registers_save (& scheduler -> threadCurrent -> registers);

      scheduler_yield (scheduler, scheduler_choose (scheduler));

      thread_registers_load (& scheduler -> threadCurrent -> registers);
        */
   }
}
   
Bool
__scheduler_enter__ (void)
{
   thread_registers_initialize ();

   if (scheduler_create (& schedulerDefault) == FALSE)
     return FALSE;

   return TRUE;
}

Bool
scheduler_create (Scheduler * scheduler)
{
   Size priority;
   TimeOffset period;

   if (schedulerCurrent != NULL)
     return FALSE;

   for (priority = THREAD_PRIORITY_IDLE;
        priority <= THREAD_PRIORITY_EXCLUSIVE;
	++ priority)
      list_clear (& scheduler -> priorities [priority]);
   
   scheduler -> priorityCurrent = & scheduler -> priorities [THREAD_PRIORITY_IDLE];
   scheduler -> threadCurrent   = NULL;

   period.seconds     = 0;
   period.nanoSeconds = 1000000000 / ((THREAD_QUANTUM_MAXIMUM - THREAD_QUANTUM_MINIMUM + 1) * 
                                      THREAD_YIELD_FREQUENCY);
   
   if (timer_create (& scheduler -> timer,
		     & period,
		     scheduler_timer_handler,
		     scheduler,
		     TRUE) == FALSE)
     return FALSE;

   schedulerCurrent = scheduler;

   if (thread_create (NULL,
                      THREAD_PRIORITY_IDLE,
                      THREAD_PRIORITY_IDLE,
                      scheduler_idle_handler,
                      scheduler,
                      NULL,
                      IDLE_THREAD_STACK_SIZE) == FALSE)
   {
      timer_destroy (& scheduler -> timer);

      return FALSE;
   }

   return TRUE;
}

Bool
scheduler_prioritize (Scheduler * scheduler)
{
   ThreadPriority oldPriority,
                  newPriority;

   if (list_empty (scheduler -> priorityCurrent) == FALSE)
     return TRUE;

   oldPriority = scheduler -> priorityCurrent - scheduler -> priorities;

   do -- scheduler -> priorityCurrent;
   while (scheduler -> priorityCurrent >= & scheduler -> priorities [THREAD_PRIORITY_IDLE] &&
	  list_empty (scheduler -> priorityCurrent) == TRUE);

   if (scheduler -> priorityCurrent < & scheduler -> priorities [THREAD_PRIORITY_IDLE])
   {
      scheduler -> priorityCurrent = & scheduler -> priorities [THREAD_PRIORITY_IDLE];

      return FALSE;
   }

   newPriority = scheduler -> priorityCurrent - scheduler -> priorities;

   if (oldPriority >= THREAD_PRIORITY_EXCLUSIVE &&
       newPriority > THREAD_PRIORITY_IDLE)
     timer_start (& scheduler -> timer);
   else
   if (oldPriority < THREAD_PRIORITY_EXCLUSIVE &&
       newPriority <= THREAD_PRIORITY_IDLE)
     timer_stop (& scheduler -> timer, NULL);

   return TRUE;
}

Bool
scheduler_preempt (Scheduler * scheduler, ThreadPriority newPriority)
{
   ThreadPriority oldPriority = scheduler -> priorityCurrent - scheduler -> priorities;
   
   if (newPriority <= oldPriority)
     return FALSE;

   if (oldPriority > THREAD_PRIORITY_IDLE &&
       newPriority >= THREAD_PRIORITY_EXCLUSIVE)
     timer_stop (& scheduler -> timer, 0);

   scheduler -> priorityCurrent = & scheduler -> priorities [newPriority];
  
   if (oldPriority <= THREAD_PRIORITY_IDLE &&
       newPriority < THREAD_PRIORITY_EXCLUSIVE)
     timer_start (& scheduler -> timer);

   if (interrupt_level () > 0 ||
       exception_level () > 0)
     thread_yield (NULL);

   return TRUE;
}

Thread *
scheduler_choose (Scheduler * scheduler)
{
   if (scheduler -> priorityCurrent != NULL)
     return (Thread *) list_front (scheduler -> priorityCurrent);

   return NULL;
}

void
scheduler_yield (Scheduler * scheduler, Thread * destination)
{
   Thread * source = scheduler -> threadCurrent;
    
   if (source != NULL &&
       source -> state == THREAD_STATE_RUNNING)
     source -> state = THREAD_STATE_READY;

   list_move_to_back (& scheduler -> priorities [destination -> priorityCurrent],
                      & destination -> threadList);

   destination -> state          = THREAD_STATE_RUNNING;
   destination -> quantumCurrent = destination -> quantumInitial;
   scheduler -> threadCurrent    = destination;
}

