#include <memory/memory.h>
#include <interrupt/interrupt.h>
#include <interrupt/exception.h>
#include <scheduler/scheduler.h>
#include <scheduler/critical.h>
#include <scheduler/registers.h>
#include <scheduler/thread.h>

Bool
thread_create (Thread ** threadResult,
	       ThreadPriority priorityFloor,
	       ThreadPriority priorityCeiling,
	       ThreadEnterFunction * enterFunction,
	       void * enterData,
               void * stack,
	       Size stackSize)
{
     Scheduler * scheduler;
     Thread * thread;

     if (priorityFloor < THREAD_PRIORITY_IDLE ||
	 priorityCeiling > THREAD_PRIORITY_EXCLUSIVE)
       return FALSE;

     scheduler = scheduler_current ();
     
     thread = (Thread *) memory_allocate (sizeof (Thread));

     if (thread == NULL)
       return FALSE;

     thread -> stackAllocated = FALSE;
     if (stack == NULL)
     {
	thread -> stackAllocated = TRUE;

	if (stackSize <= 0)
          stackSize = THREAD_STACK_SIZE;

	stack = memory_allocate (stackSize);
	if (stack == NULL)
	{
	   memory_free (thread);

	   return FALSE;
	}
     }

     thread -> scheduler       = scheduler;
     thread -> state           = THREAD_STATE_CREATE;
     
     if (priorityCeiling >= THREAD_PRIORITY_MINIMUM &&
         priorityCeiling <= THREAD_PRIORITY_MAXIMUM)
       thread -> quantumInitial = THREAD_QUANTUM_MAXIMUM - THREAD_QUANTUM_DELTA * (priorityCeiling - THREAD_PRIORITY_MINIMUM);
     else
       thread -> quantumInitial = 0;

     thread -> quantumCurrent  = thread -> quantumInitial;
     thread -> priorityFloor   = priorityFloor;
     thread -> priorityCeiling = priorityCeiling;
     thread -> priorityCurrent = priorityCeiling;

     thread -> enterFunction   = enterFunction;
     thread -> enterData       = enterData;

     thread -> stack           = (uint8 *) stack;
     thread -> stackSize       = stackSize;
  
     thread_registers_create (& thread -> registers, thread);
     
     thread_start (thread);

     if (threadResult != NULL)
       * threadResult = thread;
     
     return TRUE;
}

void *
thread_get_data (Thread * thread)
{
   if (thread == NULL)
     thread = thread_current ();
 
   return thread -> enterData;
}

void
thread_set_data (Thread * thread, void * value)
{

   if (thread == NULL)
     thread = thread_current ();

   thread -> enterData = value;
}

ThreadPriority
thread_get_priority (Thread * thread)
{
   if (thread == NULL)
     thread = thread_current ();

   return thread -> priorityCurrent;
}

ThreadPriority
thread_get_priority_floor (Thread * thread)
{
   if (thread == NULL)
     thread = thread_current ();

   return thread -> priorityFloor;
}

ThreadPriority
thread_get_priority_ceiling (Thread * thread)
{
   if (thread == NULL)
     thread = thread_current ();

   return thread -> priorityCeiling;
}

void
thread_set_priority_floor (Thread * thread, ThreadPriority floor)
{
   critical_enter ();

   if (thread == NULL) 
     thread = thread_current ();

   thread -> priorityFloor = floor;
   
   if (floor > thread -> priorityCeiling)
   {
      thread -> priorityCeiling = floor;

      thread_set_priority (thread, floor);
   }
   else
   if (thread -> priorityCurrent < floor)
     thread_set_priority (thread, floor);

   critical_exit ();
}

void
thread_set_priority_ceiling (Thread * thread, ThreadPriority ceiling)
{
   critical_enter ();

   if (thread == NULL)
     thread = thread_current ();

   thread -> priorityCeiling = ceiling;

   if (ceiling < thread -> priorityFloor)
   {
      thread -> priorityFloor = ceiling;

      thread_set_priority (thread, ceiling);
   }
   else
   if (thread -> priorityCurrent > ceiling)
     thread_set_priority (thread, ceiling);

   critical_exit ();
}

void
thread_destroy (Thread * thread)
{
   critical_enter ();

   if (thread == NULL)
     thread = thread_current ();

   if (thread -> state == THREAD_STATE_READY ||
       thread -> state == THREAD_STATE_RUNNING)
     list_remove (& thread -> threadList);

   thread_registers_destroy (& thread -> registers, thread);

   if (thread -> state == THREAD_STATE_RUNNING)
   {
       scheduler_prioritize (thread -> scheduler);

       if (thread -> stackAllocated == TRUE)
	 memory_free (thread -> stack);

       thread -> scheduler -> threadCurrent = NULL;

       memory_free (thread);

       thread_yield (NULL);
   }
   else
     memory_free (thread);

   critical_exit ();
}
   
void
thread_start (Thread * thread)
{
   critical_enter ();
  
   if (thread -> state == THREAD_STATE_CREATE ||
       thread -> state >= THREAD_STATE_STOPPED)
   {
      List * priority = & thread -> scheduler -> priorities [thread -> priorityCurrent];
      
      thread -> state = THREAD_STATE_READY;

      list_insert_back (priority, thread);

      scheduler_preempt (thread -> scheduler, thread -> priorityCurrent);
   }

   critical_exit ();
}

void
thread_stop (Thread * thread)
{
   thread_block (thread, THREAD_STATE_STOPPED, NULL, NULL);
}

void
thread_block (Thread * thread, ThreadState state, Spinlock * lock, List * list)
{
   critical_enter ();

   if (thread == NULL)
     thread = thread_current ();

   if (thread -> state == THREAD_STATE_READY ||
       thread -> state == THREAD_STATE_RUNNING)
   {
      list_remove (& thread -> threadList);

      if (lock != NULL)
        spinlock_unlock (lock);

      if (list != NULL)
      {
         ListIterator position;

	 for (position = list_begin (list);
	      position != list_end (list);
	      position = list_next (position))
	   if (thread -> priorityCurrent > ((Thread *) position) -> priorityCurrent)
	     break;
	 
         list_insert (position, & thread -> threadList);
      }

      if (thread -> state == THREAD_STATE_RUNNING)
      {
	 thread -> state = state;

	 if (thread -> priorityCurrent < thread -> priorityCeiling)
           thread_set_priority (thread, thread -> priorityCurrent + 1);

         scheduler_prioritize (thread -> scheduler);

	 thread_yield (NULL);
      }
      else
	thread -> state = state;
    
      if (lock != NULL)
        spinlock_lock (lock);
   }

   critical_exit ();
}

Bool
thread_set_priority (Thread * thread, ThreadPriority priority)
{
   critical_enter ();

   if (thread == NULL)
     thread = thread_current ();

   if (priority < thread -> priorityFloor ||
       priority > thread -> priorityCeiling)
   {
      critical_exit ();

      return FALSE;
   }

   thread -> priorityCurrent = priority;

   if (priority >= THREAD_PRIORITY_MINIMUM &&
       priority <= THREAD_PRIORITY_MAXIMUM)
     thread -> quantumInitial = THREAD_QUANTUM_MAXIMUM - THREAD_QUANTUM_DELTA * (priority - THREAD_PRIORITY_MINIMUM);
   else
     thread -> quantumInitial = 0;

   if (thread -> state == THREAD_STATE_READY ||
       thread -> state == THREAD_STATE_RUNNING)
   {
       List * priorityList = & thread -> scheduler -> priorities [priority];

       list_move_to_back (priorityList, & thread -> threadList);

       if (priority > thread -> priorityCurrent)
         scheduler_preempt (thread -> scheduler, priority);
       else
	 scheduler_prioritize (thread -> scheduler);
   }

   critical_exit ();

   return TRUE;
}
      
void
thread_yield (Thread * destination)
{
   Thread * source;
   Scheduler * scheduler;

   critical_enter ();

   scheduler = scheduler_current ();

   source = scheduler -> threadCurrent;
   
   if (destination == NULL ||
       destination -> state != THREAD_STATE_READY ||
       destination -> scheduler != scheduler)
     destination = scheduler_choose (scheduler);

   if (interrupt_level () > 0 ||
       exception_level () > 0)
   {
      if (source != NULL)
        thread_registers_save (& source -> registers);

      scheduler_yield (scheduler, destination);

      thread_registers_load (& destination -> registers);
   }
   else
   {
      scheduler_yield (scheduler, destination);
 
      if (source != NULL)
        thread_registers_switch (& source -> registers, & destination -> registers);
      else
        thread_registers_switch (NULL, & destination -> registers);
   }

   critical_exit ();
}

