#include <memory/memory.h>
#include <storage/transfer_request.h>
#include <storage/type.h>
#include <storage/storage.h>

Bool
storage_create (Storage ** result, 
                StorageType * type, 
                Size blockSize,
                StorageCapacity blockCapacity,
                void * data)
{
        Storage * storage = (Storage *) memory_allocate (sizeof (Storage));

        if (storage == NULL)
          return FALSE;
        
        mutex_create (& storage -> lock);

        storage -> type          = type;
        storage -> blockSize     = blockSize;
        storage -> blockCapacity = blockCapacity;
        storage -> state         = STORAGE_STATE_READY;
        
        list_clear (& storage -> transferRequests);

        storage -> data          = data;
        
        * result = storage;
        
        return TRUE;
}

void
storage_destroy (Storage * storage)
{
        mutex_lock (& storage -> lock);
       
        storage_cancel (storage);

        (* storage -> type -> destroy) (storage);

        mutex_destroy (& storage -> lock);

        memory_free (storage);
}
        
void
storage_cancel (Storage * storage)
{
        
        mutex_lock (& storage -> lock);

        if (storage -> state == STORAGE_STATE_BUSY)
        {
           storage -> state = STORAGE_STATE_CANCEL;

           (* storage -> type -> cancel) (storage, 
                                          (TransferRequest *) list_front (& storage -> transferRequests));

           storage -> state = STORAGE_STATE_READY;
        }

        while (list_empty (& storage -> transferRequests) == FALSE)
           transfer_request_cancel ((TransferRequest *) list_front (& storage -> transferRequests));

        mutex_unlock (& storage -> lock);
}

void
storage_flush (Storage * storage)
{
        mutex_lock (& storage -> lock);

        if (storage -> state != STORAGE_STATE_READY ||
            list_empty (& storage -> transferRequests) == TRUE)
        {
           mutex_unlock (& storage -> lock);

           return FALSE;
        }

        storage -> state = STORAGE_STATE_BUSY;

        mutex_unlock (& storage -> lock);

        (* storage -> type -> transferRequest) (storage, 
                                                (TransferRequest *) list_front (& storage -> transferRequests));

        return TRUE;
}

Bool
storage_read (Storage * storage,
              StorageBlock blockStart,
              void * buffer,
              Size bufferSize,
              TransferRequestSuccessFunction * successFunction,
              void * successData,
              TransferRequestFailureFunction * failureFunction,
              void * failureData)
{
        return transfer_request_create (NULL,
                                        storage,
                                        TRANSFER_REQUEST_TYPE_READ,
                                        blockStart,
                                        (uint8 *) buffer,
                                        bufferSize,
                                        successFunction,
                                        successData,
                                        failureFunction,
                                        failureData);
}

Bool
storage_write (Storage * storage,
               StorageBlock blockStart,
               const void * buffer,
               Size bufferSize,
               TransferRequestSuccessFunction * successFunction,
               void * successData,
               TransferRequestFailureFunction * failureFunction,
               void * failureData)
{
        return transfer_request_create (NULL,
                                        storage,
                                        TRANSFER_REQUEST_TYPE_WRITE,
                                        blockStart,
                                        (uint8 *) buffer,
                                        bufferSize,
                                        successFunction,
                                        successData,
                                        failureFunction,
                                        failureData);
}


