#include <scheduler/critical.h>
#include <timer/pit.h>
#include <timer/timer.h>

static List		  timers;
static TimeOffset	  timeCurrent;

Bool
__timer_enter__ (void)
{
   list_clear (& timers);

   timeCurrent.seconds     = 0;
   timeCurrent.nanoSeconds = 0;

   pit_start (TIMER_HZ, timer_check);

   return TRUE;
}

void
__timer_exit__ (void)
{ 
   pit_stop ();
  
   while (list_empty (& timers) == FALSE)
     timer_destroy ((Timer *) list_front (& timers));
}

void
timer_check (void)
{
   Timer * timerCurrent;

   timeCurrent.nanoSeconds += TIMER_PERIOD;
   if (timeCurrent.nanoSeconds >= NANOSECONDS_PER_SECOND)
   {
       timeCurrent.nanoSeconds -= NANOSECONDS_PER_SECOND;
       ++ timeCurrent.seconds;
   }

   for (timerCurrent = (Timer *) list_front (& timers);
        list_empty (& timers) == FALSE;
        timerCurrent = (Timer *) list_front (& timers))
   {
      if (timerCurrent -> timeDeadline.seconds > timeCurrent.seconds)
        break;

      if (timerCurrent -> timeDeadline.seconds < timeCurrent.seconds ||
          timerCurrent -> timeDeadline.nanoSeconds <= timeCurrent.nanoSeconds)
        timer_activate (timerCurrent);
      else
        break;
   }
}

Bool
timer_start (Timer * timer)
{
   ListIterator position;

   critical_enter ();

   if (timer -> timeDeadline.seconds != 0 ||
       timer -> timeDeadline.nanoSeconds != 0)
   {
      critical_exit ();

      return FALSE;
   }

   timer -> timeDeadline = timeCurrent;
   time_offset_add (& timer -> timeDeadline, & timer -> timeOffset);

   for (position = list_begin (& timers);
        position != list_end (& timers);
	position = list_next (position))
   {
       Timer * currentTimer = (Timer *) position;

       if (timer -> timeDeadline.seconds > currentTimer -> timeDeadline.seconds)
	 continue;

       if (timer -> timeDeadline.seconds < currentTimer -> timeDeadline.seconds ||
	   timer -> timeDeadline.nanoSeconds < currentTimer -> timeDeadline.nanoSeconds)
	 break;
   }

   list_insert (position, & timer -> timerList);

   critical_exit ();

   return TRUE;
}

Bool
timer_stop (Timer * timer, TimeOffset * remaining)
{
   critical_enter ();

   if (timer -> timeDeadline.seconds == 0 &&
       timer -> timeDeadline.nanoSeconds == 0)
   {
       if (remaining != NULL)
       {
          remaining -> seconds     = 0;
          remaining -> nanoSeconds = 0;
       }

       critical_exit ();

       return FALSE;
   }
   
   list_remove (& timer -> timerList);

   if (remaining != NULL)
   {
       * remaining = timer -> timeDeadline;
       time_offset_subtract (remaining, & timeCurrent);
   }

   timer -> timeDeadline.seconds     = 0;
   timer -> timeDeadline.nanoSeconds = 0;
   
   critical_exit ();

   return TRUE;
}

Bool 
timer_reset (Timer * timer, TimeOffset * remaining)
{   
   ListIterator position;

   critical_enter ();

   if (timer -> timeDeadline.seconds == 0 &&
       timer -> timeDeadline.nanoSeconds == 0)
   {
      if (remaining != NULL)
      {
	 remaining -> seconds     = 0;
	 remaining -> nanoSeconds = 0;
      }

      critical_exit ();

      return FALSE;
   }
   
   if (remaining != NULL)
   {
      * remaining = timer -> timeDeadline;
      time_offset_subtract (remaining, & timeCurrent);
   }

   timer -> timeDeadline = timeCurrent;
   time_offset_add (& timer -> timeDeadline, & timer -> timeOffset);

   for (position = list_next (& timer -> timerList);
        position != list_end (& timers);
        position = list_next (position))
   {
       Timer * currentTimer = (Timer *) position;

       if (timer -> timeDeadline.seconds > currentTimer -> timeDeadline.seconds)
	 continue;

       if (timer -> timeDeadline.seconds < currentTimer -> timeDeadline.seconds ||
           timer -> timeDeadline.nanoSeconds < currentTimer -> timeDeadline.nanoSeconds)
         break;
   }

   list_remove (& timer -> timerList);
   list_insert (position, & timer -> timerList);

   critical_exit ();

   return TRUE;
}

void
timer_activate (Timer * timer)
{
   critical_enter ();

   if (timer -> isPeriodic == TRUE)
     timer_reset (timer, NULL);
   else
     timer_stop (timer, NULL);

   (* timer -> handlerFunction) (timer, timer -> handlerData);

   critical_exit ();
}

Bool
timer_create (Timer * timer, 
              const TimeOffset * timeOffset,
	      TimerHandler * handlerFunction,
	      void * handlerData,
	      Bool isPeriodic)
{
     NanoSeconds lostPrecision;

     if (timer == NULL)
       return FALSE;

     timer -> timeDeadline.seconds     = 0;
     timer -> timeDeadline.nanoSeconds = 0;

     timer -> timeOffset = * timeOffset;
            
     time_offset_normalize (& timer -> timeOffset);
     
     lostPrecision = timer -> timeOffset.nanoSeconds % TIMER_PERIOD;
     if (lostPrecision > 0)
     {
         timer -> timeOffset.nanoSeconds -= lostPrecision;
	 timer -> timeOffset.nanoSeconds += TIMER_PERIOD;
     }

     timer -> handlerFunction = handlerFunction;
     timer -> handlerData     = handlerData;
     
     timer -> isPeriodic = isPeriodic;

     return TRUE;
}

void
timer_destroy (Timer * timer)
{
     critical_enter ();
     
     if (timer -> timeDeadline.seconds > 0 ||
         timer -> timeDeadline.nanoSeconds > 0)
       timer_stop (timer, NULL);

     critical_exit ();
}

