#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "elf.h"
#include "xcom.h"
#include "error.h"
#include "object.h"
#include "symbol.h"
#include "section.h"
#include "relocation.h"

int
elf_verify (FILE * file)
{
        ELFHeader header;

        rewind (file);

        if (fread (& header, sizeof (ELFHeader), 1, file) != 1 ||
            memcmp (header.signature, ELF_SIGNATURE, ELF_SIGNATURE_LENGTH) != 0)
          return 0;

        return 1;
}
        
int
elf_read (Object * object, FILE * file)
{
        ELFHeader header;
        ELFSectionHeader * sectionMap = NULL,
                         * currentSection;
        Section * stringSection       = NULL;
        Section ** objectSections     = NULL;

        rewind (file);

        if (fread (& header, sizeof (ELFHeader), 1, file) != 1)
          xl_error ("Could not read ELF header");

        if (memcmp (header.signature, ELF_SIGNATURE, ELF_SIGNATURE_LENGTH) != 0)
          xl_error ("Invalid signature in ELF header");

        if (header.fileClass != ELF_HOST_FILE_CLASS)
          xl_error ("Unsupported file class in ELF header");

        if (header.dataEncoding != ELF_HOST_DATA_ENCODING)
          xl_error ("Unsupported data encoding in ELF header");

        sectionMap = (ELFSectionHeader *) malloc (header.sectionHeaderCount * sizeof (ELFSectionHeader));
        if (sectionMap == NULL)
          goto failure;

        objectSections = (Section **) calloc (header.sectionHeaderCount, sizeof (Section *));
        if (objectSections == NULL)
          goto failure;

        if (fseek (file, header.sectionHeaderOffset, SEEK_SET) != 0)
          goto failure;

        if (fread (sectionMap, header.sectionHeaderCount * sizeof (ELFSectionHeader), 1, file) != 1)
          xl_error ("Could not read section headers from ELF");

        stringSection = section_create (object, SECTION_TYPE_TEMPORARY, 0, 0, 0, 0, 0, 0, (size_t) sectionMap [header.stringSection].size, NULL);
        if (stringSection == NULL)
          goto failure;

        objectSections [header.stringSection] = stringSection;

        if (fseek (file, sectionMap [header.stringSection].offset, SEEK_SET) != 0 ||
            section_read (stringSection, file) != 0)
          xl_error ("Could not read string section from ELF");

        for (currentSection = sectionMap;
             currentSection < & sectionMap [header.sectionHeaderCount];
             ++ currentSection)
        {
            Section * newSection;
            SectionType type = SECTION_TYPE_TEMPORARY;
            const char * name = NULL,
                       * architecture = NULL;
            SectionAttributes attributes = 0;
            SectionCapabilities capabilities = 0;
            size_t alignment;

            if (objectSections [currentSection - sectionMap] != NULL)
              continue;

            switch (currentSection -> type)
            {
            case ELF_SECTION_TYPE_PROGRAM:
            case ELF_SECTION_TYPE_ALLOCATE:
                 if ((currentSection -> attributes & ELF_SECTION_ATTRIBUTE_LOAD) == 0)
                   continue;

                 type         = SECTION_TYPE_PROGRAM;
                 attributes   = SECTION_ATTRIBUTE_MANDATORY;
                 capabilities = SECTION_CAPABILITY_READ;

                 if (currentSection -> type == ELF_SECTION_TYPE_PROGRAM)
                   attributes |= SECTION_ATTRIBUTE_CONTENTS;

                 if (currentSection -> address > 0)
                   attributes |= SECTION_ATTRIBUTE_ANCHORED;

                 if (currentSection -> attributes & ELF_SECTION_ATTRIBUTE_WRITE)
                   capabilities |= SECTION_CAPABILITY_WRITE;

                 if (currentSection -> attributes & ELF_SECTION_ATTRIBUTE_EXECUTE)
                 {
                    capabilities |= SECTION_CAPABILITY_EXECUTE;
  
                    architecture = XCOM_HOST_ARCHITECTURE_CODE;
                 }
                 else
                   architecture = XCOM_HOST_ARCHITECTURE_DATA;
               
                 break;

            case ELF_SECTION_TYPE_NOTE:
                 type = SECTION_TYPE_RESOURCE;

                 break;
          
            case ELF_SECTION_TYPE_STRINGS: 
            case ELF_SECTION_TYPE_SYMBOLS:
            case ELF_SECTION_TYPE_RELOCATIONS:
            case ELF_SECTION_TYPE_ADDEND_RELOCATIONS:
                 break;

            default:
                 continue;
            }

            for (alignment = 0;
                 (size_t) (1 << alignment) < currentSection -> alignment;
                 ++ alignment);

            if (currentSection -> name != 0)
              name = (char *) & stringSection -> data [currentSection -> name];

            newSection = object_define_section (object,
                                                type,
                                                attributes,
                                                capabilities,
                                                name,
                                                architecture,
                                                (Address) currentSection -> address,
                                                alignment,
                                                (size_t) currentSection -> size);

            objectSections [currentSection - sectionMap] = newSection;

            if (currentSection -> type != ELF_SECTION_TYPE_ALLOCATE)
            {
               if (fseek (file, currentSection -> offset, SEEK_SET) != 0 ||
                   section_read (newSection, file) != 0)
                 xl_error ("Failed to read section contents from ELF");
            }
        }

        for (currentSection = sectionMap;
             currentSection < & sectionMap [header.sectionHeaderCount];
             ++ currentSection)
        {
            Section * targetSection,
                    * nameSection;
            ELFSymbol * currentSymbol,
                      * endSymbol;
            const char * name;

            if (currentSection -> type != ELF_SECTION_TYPE_SYMBOLS)
              continue;

            nameSection = objectSections [currentSection -> link];
            
            currentSymbol = (ELFSymbol *) objectSections [currentSection - sectionMap] -> data;
            endSymbol     = & currentSymbol [currentSection -> size / sizeof (ELFSymbol)];
            
            while (currentSymbol < endSymbol)
            {
                SymbolType type;
                SymbolAttributes attributes;

                if (currentSymbol -> name == 0)
                {
                   ++ currentSymbol;
                   
                   continue;
                }

                switch (currentSymbol -> attributes & ELF_SYMBOL_ATTRIBUTE_TYPE)
                {
                case ELF_SYMBOL_TYPE_DATA:
                     type = SYMBOL_TYPE_DATA;

                     break;

                case ELF_SYMBOL_TYPE_FUNCTION:
                     type = SYMBOL_TYPE_FUNCTION;

                     break;

                case ELF_SYMBOL_TYPE_SECTION:
                     type = SYMBOL_TYPE_SECTION;

                     break;

                case ELF_SYMBOL_TYPE_FILE:
                     type = SYMBOL_TYPE_FILE;

                     break;

                default:
                     type = SYMBOL_TYPE_GENERIC;

                     break;
                }

                switch (currentSymbol -> attributes & ELF_SYMBOL_ATTRIBUTE_BINDING)
                {
                case ELF_SYMBOL_BINDING_GLOBAL:
                case ELF_SYMBOL_BINDING_WEAK:
                     attributes = SYMBOL_ATTRIBUTE_EXPORTED | SYMBOL_ATTRIBUTE_MANDATORY;

                     break;

                default:
                     attributes = 0; 
                }

                targetSection = NULL;

                name = (char *) & nameSection -> data [currentSymbol -> name];

                switch (currentSymbol -> section)
                {
                case ELF_SECTION_INDEX_UNDEFINED:
                     attributes |= SYMBOL_ATTRIBUTE_IMPORTED;

                     break;

                case ELF_SECTION_INDEX_ABSOLUTE:
                     if (type != SYMBOL_TYPE_FILE)
                       type = SYMBOL_TYPE_CONSTANT;

                     break;

                case ELF_SECTION_INDEX_COMMON:
                     attributes |= SYMBOL_ATTRIBUTE_COMMON;
                        
                     break;

                default:
                     targetSection = objectSections [currentSymbol -> section];
                        
                     break;
                }

                if (type != SYMBOL_TYPE_SECTION)
                  object_define_symbol (object,
                                        targetSection,
                                        type,
                                        attributes,
                                        name,
                                        (Address) currentSymbol -> value,
                                        (size_t) currentSymbol -> size);
               
                ++ currentSymbol;
            }
        }

        for (currentSection = sectionMap;
             currentSection < & sectionMap [header.sectionHeaderCount];
             ++ currentSection)
        {
            Section * targetSection,
                    * symbolSection;
            uint8 * currentReloc,
                  * endReloc;

            if (currentSection -> type != ELF_SECTION_TYPE_RELOCATIONS &&
                currentSection -> type != ELF_SECTION_TYPE_ADDEND_RELOCATIONS)
              continue;

            if (sectionMap [currentSection -> information].type != ELF_SECTION_TYPE_PROGRAM)
              xl_error ("Relocations for non-program section in ELF");

            targetSection = objectSections [currentSection -> information];
            symbolSection = objectSections [currentSection -> link];

            currentReloc = objectSections [currentSection - sectionMap] -> data;
            endReloc     = & currentReloc [currentSection -> size];

            while (currentReloc < endReloc)
            {
                RelocationType type = 0;
                ELFSymbol * elfSymbol;
                ELFRelocation * elfReloc;
                const char * name;
                Symbol * symbol;
              
                elfReloc = (ELFRelocation *) currentReloc;
                
                if (elfReloc -> symbol == 0)
                  xl_error ("Unsupported relocation type in ELF");

                elfSymbol = (ELFSymbol *) symbolSection -> data + elfReloc -> symbol;
                
                if (elfSymbol -> name == 0)
                  name = objectSections [elfSymbol -> section] -> name;
                else
                  name = (char *) & objectSections [sectionMap [currentSection -> link].link] -> data [elfSymbol -> name];
                
                symbol = object_find_symbol (object, name);
                if (symbol == NULL)
                  xl_error ("Relocation against unknown symbol %s in ELF", name);

                switch (elfReloc -> type)
                {
                case ELF_RELOCATION_TYPE_I386_ABSOLUTE_32:
                     type = RELOCATION_TYPE_ABSOLUTE_32;

                     break;

                case ELF_RELOCATION_TYPE_I386_PC_RELATIVE_32:
                     type = RELOCATION_TYPE_RELATIVE_32;

                     break;
                
                default:
                     xl_error ("Unsupported relocation type in ELF");
                }
       
                if (currentSection -> type == ELF_SECTION_TYPE_ADDEND_RELOCATIONS)
                {
                     elf_int32 addend = ((ELFAddendRelocation *) currentReloc) -> addend;

                     if (addend > 0)
                       * (elf_uint32 *) & targetSection -> data [elfReloc -> offset] += addend;
                }

                if (relocation_create (& targetSection -> relocations,
                                       type,
                                       (size_t) elfReloc -> offset,
                                       symbol -> definition) == NULL)
                  goto failure;
                
                currentReloc += currentSection -> entrySize;
            }
        }
       
        for (currentSection = sectionMap;
             currentSection < & sectionMap [header.sectionHeaderCount];
             ++ currentSection)
        {
            if (currentSection -> type == ELF_SECTION_TYPE_STRINGS ||
                currentSection -> type == ELF_SECTION_TYPE_SYMBOLS ||
                currentSection -> type == ELF_SECTION_TYPE_RELOCATIONS ||
                currentSection -> type == ELF_SECTION_TYPE_ADDEND_RELOCATIONS)
              section_destroy (objectSections [currentSection - sectionMap]);
        }

        free (sectionMap);
        free (objectSections);

        return 0;

failure:
        if (stringSection != NULL)
          section_destroy (stringSection);

        if (sectionMap != NULL)
          free (sectionMap);

        if (objectSections != NULL)
          free (objectSections);

        return -1;
}

