#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "error.h"
#include "symbol.h"
#include "object.h"

void
xl_version (void)
{
    printf ("XL - XCOM Linker (%s) by Lee Salzman (eihrul@tunes.org)\n", __DATE__);
}

void
xl_help (void)
{
    printf ("Usage: xl [options] output input1 ... inputN\n"
            "\n"
            "Options:\n"
            "   -h                      Display help information\n"
            "   -v                      Display version information\n"
            "   -m                      Display verbose messages\n"                 
            "   -z                      Zero-fill allocated sections\n"
            "   -s                      Strip extraneous sections and symbols\n"
            "   -e                      Keep exported symbols from becoming stripped\n"
            "   -c section              Allocate common symbols in section\n"
            "   -a address              Anchor output file at specified address\n"
            "   -a address@section      Anchor section in output file at specified address\n"
            "   -i interface@type       Link in the interface file as the specified type\n"
            "\n");
}    

int
main (int argc, char ** argv)
{
    FILE * input, * output;
    Object destination, source;
    int argumentCurrent,
        argumentCount,
        stripExtraneous = 0,
        keepExportedSymbols = 0,
        zeroFill = 0;
    
    for (argumentCurrent = 1,
           argumentCount = argc;
         argumentCurrent < argumentCount;
         ++ argumentCurrent)
    {
       if (argv [argumentCurrent] [0] == '-')
       {
           int switchArguments = 1;
           char * switchCurrent;

           for (switchCurrent = & argv [argumentCurrent] [1];
                * switchCurrent != '\0';
                ++ switchCurrent)
           switch (* switchCurrent)
           {
           case 'v':
               xl_version ();

               return EXIT_SUCCESS;

           case 'm':
               xl_verbose (1);

               break;
          
           case 's':
               stripExtraneous = 1;

               break;

           case 'z':
               zeroFill = 1;

               break;

           case 'e':
               keepExportedSymbols = 1;

               break;

           case 'a':
               switchArguments += 1;

               break;

           case 'c':
               switchArguments += 1;

               break;

           case 'i':
               switchArguments += 1;

               break;

           case 'h':
               xl_help ();

               return EXIT_SUCCESS;

           default:
               xl_help ();

               return EXIT_FAILURE;
           }

           if (argumentCount - argumentCurrent < switchArguments)
           {
              xl_help ();

              return EXIT_FAILURE;
           }

           while (switchArguments > 0)
           {
               char * switchArgument = argv [argumentCurrent];

               memcpy (& argv [argumentCurrent], 
                       & argv [argumentCurrent + 1], 
                       (argc - argumentCurrent - 1) * sizeof (char *));

               argv [argc - 1] = switchArgument;

               -- switchArguments;
               -- argumentCount;
           }
          
           -- argumentCurrent;
       }
    }
    
    if (argumentCount < 2)
      xl_error ("No output file specified");
    
    if (argumentCount < 3)
      xl_error ("No input files specified");

    object_create (& destination);

    xl_message ("Merging input file %s", argv [2]);
    
    input = fopen (argv [2], "rb");
    
    if (input == NULL ||
        object_read (& destination, input) < 0)
      xl_error ("Failed to read input file %s", argv [2]);

    if (stripExtraneous != 0)
      object_strip (& destination, keepExportedSymbols);

    for (argumentCurrent = 3;
         argumentCurrent < argumentCount;
         ++ argumentCurrent)
    {
        object_create (& source);

        xl_message ("Merging input file %s", argv [argumentCurrent]);

        input = freopen (argv [argumentCurrent], "rb", input);

        if (input == NULL ||
            object_read (& source, input) < 0)
          xl_error ("Failed to read input file %s", argv [argumentCurrent]);

        if (stripExtraneous != 0)
          object_strip (& source, keepExportedSymbols);

        if (object_merge (& destination, & source) < 0)
          xl_error ("Failed to merge input file %s", argv [argumentCurrent]);
    }

    fclose (input);

    for (argumentCurrent = argumentCount;
         argumentCurrent < argc;
         ++ argumentCurrent)
    {
       const char * switchCurrent;
       
       for (switchCurrent = & argv [argumentCurrent] [1];
            * switchCurrent != '\0';
            ++ switchCurrent)
       switch (* switchCurrent)
       {
       case 'a':
            { 
               const char * section;
               Address address;
               int length;
              
               ++ argumentCurrent;

               if (sscanf (argv [argumentCurrent], "%i%n", (int *) & address, & length) != 1)
                 xl_error ("Bad anchor address specified for output file %s", argv [1]);
     
               section = argv [argumentCurrent] + length;
               if (* section == '@')
                 ++ section;
               else
                 section = NULL;
       
               if (section != NULL)
                 xl_message ("Anchoring output file %s to 0x%X at section %s", argv [1], address, section);
               else
                 xl_message ("Anchoring output file %s to 0x%X", argv [1], address);
       
               if (object_anchor (& destination, address, section) < 0)
                 xl_error ("Failed to anchor output file %s", argv [1]);

               break;
            }

       case 'c':
            ++ argumentCurrent;
            
            object_allocate_common_symbols (& destination, argv [argumentCurrent]);

            break;

       case 'i':
            {
               const char * fileName,
                          * interfaceType;
               Section * section;
               
               ++ argumentCurrent;

               fileName = strtok (argv [argumentCurrent], "@");
               interfaceType = strtok (NULL, "");

               if (interfaceType == NULL)
                 xl_error ("Bad interface type specified for interface file %s", fileName);

               xl_message ("Merging interface file %s as interface type %s", fileName, interfaceType);

               input = fopen (fileName, "rb");

               if (input == NULL ||
                   fseek (input, 0, SEEK_END) != 0)
                 xl_error ("Failed to read interface file %s", fileName);

               section = section_create (& destination,
                                         SECTION_TYPE_INTERFACE,
                                         SECTION_ATTRIBUTE_MANDATORY,
                                         0,
                                         NULL,
                                         object_translate_string (& destination, interfaceType),
                                         0,
                                         0,
                                         (size_t) ftell (input),
                                         NULL);

               rewind (input);
               
               if (section == NULL ||
                   section_read (section, input) != 0)
                 xl_error ("Failed to read interface file %s", fileName);

               fclose (input);

               break;
            }
 
       default:
            break;
       }
    }

    object_relocate (& destination);

    if (stripExtraneous != 0)
      object_strip (& destination, keepExportedSymbols);

    output = fopen (argv [1], "wb");

    xl_message ("Creating output file %s", argv [1]);

    if (output == NULL ||
        object_write (& destination, output, zeroFill) < 0)
      xl_error ("Failed to write output file %s", argv [1]);

    fclose (output);

    object_destroy (& destination);

    return EXIT_SUCCESS;
}
    
