#include "xcom.h"
#include "error.h"
#include "object.h"
#include "symbol.h"
#include "section.h"
#include "relocation.h"

int
object_create (Object * object)
{
    list_clear (& object -> sections);
    list_clear (& object -> importedSymbols);
    list_clear (& object -> constantSymbols);
    list_clear (& object -> sectionSymbols);

    if (namespace_create (& object -> namespace) != 0)
      return -1;

    return 0;
}

void
object_destroy (Object * object)
{
    while (list_empty (& object -> sections) == 0)
       section_destroy ((Section *) list_front (& object -> sections));

    while (list_empty (& object -> importedSymbols) == 0)
       symbol_destroy ((Symbol *) list_front (& object -> importedSymbols));

    while (list_empty (& object -> constantSymbols) == 0)
       symbol_destroy ((Symbol *) list_front (& object -> constantSymbols));

    while (list_empty (& object -> sectionSymbols) == 0)
       symbol_destroy ((Symbol *) list_front (& object -> sectionSymbols));

    namespace_destroy (& object -> namespace);
}

int
object_anchor (Object * object, Address address, const char * name)
{
    ListIterator position;

    for (position = list_begin (& object -> sections);
         position != list_end (& object -> sections);
         position = list_next (position))
    {
       Section * section = (Section *) position;

       if (section -> type != SECTION_TYPE_PROGRAM ||
           (name != NULL &&
            strcmp (section -> name, name) != 0))
         continue;

       if (section_anchor (section, address) != address)
         return -1;

       address += section -> dataSize;
    }

    return 0;
}

int
object_merge (Object * dst, Object * src)
{
    ListIterator dstPos, srcPos;

    object_define_symbols (dst, NULL, & src -> constantSymbols);
    object_define_symbols (dst, NULL, & src -> importedSymbols);
    object_define_symbols (dst, NULL, & src -> sectionSymbols);

    for (dstPos = list_begin (& dst -> sections);
         dstPos != list_end (& dst -> sections);
         dstPos = list_next (dstPos))
    {
        Section * dstSection = (Section *) dstPos,
                * srcSection;
        
        for (srcPos = list_begin (& src -> sections);
             srcPos != list_end (& src -> sections);
             srcPos = list_next (srcPos))
        {
           srcSection = (Section *) srcPos;
         
           if (section_adjacent (dstSection, srcSection) != 0)
           {
              object_fixup (src, srcSection, dstSection -> dataSize);

              break;
           }
        }
    }

    for (dstPos = list_begin (& dst -> sections);
         dstPos != list_end (& dst -> sections);
         dstPos = list_next (dstPos))
    {
        Section * dstSection = (Section *) dstPos,
                * srcSection;

        for (srcPos = list_begin (& src -> sections);
             srcPos != list_end (& src -> sections);
             srcPos = list_next (srcPos))
        {
           srcSection = (Section *) srcPos;
         
           if (section_adjacent (dstSection, srcSection) != 0)
           {
              if (section_merge (dstSection, srcSection) != 0)
                return -1;

              break;
           }
        }
    }

    while (list_empty (& src -> sections) == 0)
      if (section_move ((Section *) list_front (& src -> sections), dst) != 0)
        return -1;
  
    object_destroy (src);

    return 0;
}

void
object_fixup (Object * object, Section * target, size_t offset)
{
     ListIterator position;

     for (position = list_begin (& object -> sections);
          position != list_end (& object -> sections);
          position = list_next (position))
     {
         Section * section = (Section *) position;

         if (section -> type != SECTION_TYPE_PROGRAM ||
             (section -> attributes & SECTION_ATTRIBUTE_CONTENTS) == 0)
           continue;

         section_fixup (section, target, offset);
     }
}

void
object_relocate (Object * object)
{
     ListIterator position;

     for (position = list_begin (& object -> sections);
          position != list_end (& object -> sections);
          position = list_next (position))
     {
         Section * section = (Section *) position;
        
         if (section -> type != SECTION_TYPE_PROGRAM ||
             (section -> attributes & SECTION_ATTRIBUTE_CONTENTS) == 0)
           continue;

         section_relocate (section);
     }
}

Symbol *
object_find_symbol (Object * object, const char * name)
{
     Definition * definition = namespace_find (& object -> namespace, name);

     if (definition == NULL)
       return NULL;

     return definition -> symbol;
}

Symbol *
object_define_symbol (Object * object,
                      Section * section,
                      SymbolType type,
                      SymbolAttributes attributes,
                      const char * name,
                      Address value,
                      size_t size)
{
     Definition * definition = namespace_find (& object -> namespace, name);
     Symbol * symbol;

     if (definition == NULL)
       definition = namespace_insert (& object -> namespace,
                                      name,
                                      NULL);
     
     symbol = definition -> symbol;

     if ((attributes & (SYMBOL_ATTRIBUTE_EXPORTED | SYMBOL_ATTRIBUTE_COMMON)) != 0 &&
         symbol != NULL)
     {
        if (type == SYMBOL_TYPE_GENERIC)
          type = symbol -> type;
        else
        if (symbol -> type != type &&
            symbol -> type != SYMBOL_TYPE_GENERIC)
          xl_error ("Differing types for symbols with the same name %s", name);
       
        if (symbol -> size != size)
        {
           if (type == SYMBOL_TYPE_SECTION)
             size += symbol -> size;
           else
           if (size == 0)
             size = symbol -> size;
           else
            if (symbol -> size != 0)
             xl_error ("Differing sizes for symbols with the same name %s", name);
        }

        if ((symbol -> attributes & SYMBOL_ATTRIBUTE_COMMON) != (attributes & SYMBOL_ATTRIBUTE_COMMON))
          xl_error ("Conflicting allocation for symbols with the same name %s", name);

        if (type != SYMBOL_TYPE_SECTION &&
            (symbol -> attributes & SYMBOL_ATTRIBUTE_IMPORTED) == 0 &&
            (attributes & SYMBOL_ATTRIBUTE_IMPORTED) == 0 &&
            (symbol -> section != section ||
             symbol -> value != value))
          xl_error ("Conflicting values for symbols with the same name %s", name);

        if (symbol -> attributes & SYMBOL_ATTRIBUTE_IMPORTED)
        {
           symbol -> type       = type;
           symbol -> attributes = attributes | (symbol -> attributes & SYMBOL_ATTRIBUTE_MANDATORY);
           symbol -> section    = section;
           symbol -> value      = value;
           symbol -> size       = size;

           if (section != NULL)
             list_move_to_front (& section -> exportedSymbols, & symbol -> symbolList);
           else
             list_move_to_front (& object -> constantSymbols, & symbol -> symbolList);
        }
        else
          symbol -> attributes |= attributes & (SYMBOL_ATTRIBUTE_EXPORTED | SYMBOL_ATTRIBUTE_MANDATORY);
     }
     else
     {
        symbol = symbol_create (section, type, attributes, definition, value, size);
        if (symbol == NULL)
          return NULL;
       
        if (type == SYMBOL_TYPE_SECTION)
          list_insert_front (& object -> sectionSymbols, & symbol -> symbolList);
        else
        if (attributes & (SYMBOL_ATTRIBUTE_IMPORTED | SYMBOL_ATTRIBUTE_COMMON))
          list_insert_front (& object -> importedSymbols, & symbol -> symbolList);
        else
        if (section == NULL)
          list_insert_front (& object -> constantSymbols, & symbol -> symbolList);

        if (attributes & (SYMBOL_ATTRIBUTE_EXPORTED | SYMBOL_ATTRIBUTE_COMMON))
          definition -> symbol = symbol;

        definition_reference (definition);
     }
        
     return symbol;
}

void
object_define_symbols (Object * object, Section * section, List * symbols)
{
     ListIterator position;

     for (position = list_begin (symbols);
          position != list_end (symbols);
          position = list_next (position))
     {
         Symbol * symbol = (Symbol *) position;

         object_define_symbol (object,
                               symbol -> type == SYMBOL_TYPE_SECTION ? symbol -> section : section,
                               symbol -> type,
                               symbol -> attributes,
                               symbol -> definition -> name,
                               symbol -> value,
                               symbol -> size);
     }
}

Section *
object_find_section (Object * object,
                     const char * name)
{
     ListIterator position;

     for (position = list_begin (& object -> sections);
          position != list_end (& object -> sections);
          position = list_next (position))
     {
         Section * section = (Section *) position;

         if (section -> name != NULL &&
             strcmp (section -> name, name) == 0)
           return section;
     }

     return NULL;
}

Section *
object_define_section (Object * object,
                       SectionType type,
                       SectionAttributes attributes,
                       SectionCapabilities capabilities,
                       const char * name,
                       const char * architecture,
                       Address address,
                       size_t alignment,
                       size_t size)
{
     Section * section;

     if (name != NULL)
       name = object_translate_string (object, name);

     if (architecture != NULL)
       architecture = object_translate_string (object, architecture);

     section = section_create (object,
                               type,
                               attributes,
                               capabilities,
                               name,
                               architecture,
                               address,
                               alignment,
                               size,
                               NULL);
     if (section == NULL)
     {
        if (name != NULL)
          xl_error ("Failed to create section %s", name);
     }

     if (type == SECTION_TYPE_PROGRAM &&
         name != NULL)
       object_define_symbol (object,
                             section,
                             SYMBOL_TYPE_SECTION,
                             SYMBOL_ATTRIBUTE_EXPORTED,
                             name,
                             0,
                             size);

     return section;
}

const char *
object_translate_string (Object * object, const char * string)
{
     Definition * definition = namespace_find (& object -> namespace, string);

     if (definition == NULL)
       definition = namespace_insert (& object -> namespace, string, NULL);

     definition_reference (definition);

     return definition -> name;
}

void
object_strip (Object * object, int keepExportedSymbols)
{
     ListIterator sectionPos,
                  symbolPos;

     symbolPos = list_begin (& object -> constantSymbols);
     
     while (symbolPos != list_end (& object -> constantSymbols))
     {
        Symbol * symbol = (Symbol *) symbolPos;

        symbolPos = list_next (symbolPos);

        if ((symbol -> attributes & SYMBOL_ATTRIBUTE_MANDATORY) != 0 ||
            symbol -> definition -> relocationCount > 0 ||
            (keepExportedSymbols != 0 &&
             (symbol -> attributes & (SYMBOL_ATTRIBUTE_EXPORTED | SYMBOL_ATTRIBUTE_COMMON)) != 0))
          continue;

        symbol_destroy (symbol);
     }
         
     sectionPos = list_begin (& object -> sections);

     while (sectionPos != list_end (& object -> sections))
     {
        Section * section = (Section *) sectionPos;

        symbolPos = list_begin (& section -> exportedSymbols);

        while (symbolPos != list_end (& section -> exportedSymbols))
        {
           Symbol * symbol = (Symbol *) symbolPos;

           symbolPos = list_next (symbolPos);

           if ((symbol -> attributes & SYMBOL_ATTRIBUTE_MANDATORY) != 0 ||
               symbol -> definition -> relocationCount > 0 ||
               (keepExportedSymbols != 0 &&
                (symbol -> attributes & (SYMBOL_ATTRIBUTE_EXPORTED | SYMBOL_ATTRIBUTE_COMMON)) != 0))
             continue;

           symbol_destroy (symbol);
        }

        sectionPos = list_next (sectionPos);

        if (section -> type != SECTION_TYPE_RESOURCE &&
            section -> type != SECTION_TYPE_INTERFACE)
          continue;

        if (section -> attributes & SECTION_ATTRIBUTE_MANDATORY)
          continue;

        section_destroy (section);
     }
}

int        
object_read (Object * object, FILE * file)
{
     if (xcom_verify (file) != 0)
       return xcom_read (object, file);

     if (elf_verify (file) != 0)
       return elf_read (object, file);

     xl_error ("Unknown object file format");

     return -1;
}

int
object_write (Object * object, FILE * file, int zeroFill)
{
     return xcom_write (object, file, zeroFill);
}

void
object_allocate_common_symbols (Object * object, const char * sectionName)
{
     Section * section = object_find_section (object, sectionName);
     ListIterator position;

     if (section != NULL)
     {
        if (section -> attributes & SECTION_ATTRIBUTE_CONTENTS)
          xl_error ("Can't allocate common symbols in section %s with contents", sectionName);
     }
     else
       section = object_define_section (object,
                                        SECTION_TYPE_PROGRAM,
                                        SECTION_ATTRIBUTE_MANDATORY,
                                        SECTION_CAPABILITY_READ | SECTION_CAPABILITY_WRITE,
                                        sectionName,
                                        XCOM_HOST_ARCHITECTURE_DATA,
                                        0,
                                        0,
                                        (size_t) XCOM_HOST_ALIGNMENT_DATA);

     position = list_begin (& object -> importedSymbols);

     while (position != list_end (& object -> importedSymbols))
     {
         Symbol * symbol = (Symbol *) position;

         position = list_next (position);

         if ((symbol -> attributes & SYMBOL_ATTRIBUTE_COMMON) == 0)
           continue;

         symbol -> value = (Address) section -> dataSize;
        
         section -> dataSize += symbol -> size;

         symbol -> attributes &= ~ SYMBOL_ATTRIBUTE_COMMON;
         
         list_move_to_back (& section -> exportedSymbols, & symbol -> symbolList);
     }
}

