#include <stdlib.h>
#include <string.h>
#include "xcom.h"
#include "error.h"
#include "object.h"
#include "symbol.h"
#include "section.h"
#include "relocation.h"

int
xcom_verify (FILE * file)
{
     XComHeader header;

     rewind (file);

     if (fread (& header, sizeof (XComHeader), 1, file) != 1 ||
         memcmp (header.signature, XCOM_SIGNATURE, XCOM_SIGNATURE_LENGTH) != 0)
       return 0;

     return 1;
}

int
xcom_read (Object * object, FILE * file)
{
     XComHeader header;
     XComSection * sectionMap  = NULL, 
                 * currentSection;
     Section * stringSection   = NULL;
     Section ** objectSections = NULL;

     rewind (file);

     if (fread (& header, sizeof (XComHeader), 1, file) != 1)
       xl_error ("Could not read XCOM header");

     if (memcmp (header.signature, XCOM_SIGNATURE, XCOM_SIGNATURE_LENGTH) != 0)
       xl_error ("Invalid signature in XCOM header");

     if ((xc_reference) (header.referenceAttributes & XCOM_REFERENCE_ATTRIBUTE_SIZE) != XCOM_HOST_REFERENCE_ATTRIBUTE_SIZE ||
         (xc_reference) (header.referenceAttributes & XCOM_REFERENCE_ATTRIBUTE_BIG_ENDIAN) != XCOM_HOST_REFERENCE_ATTRIBUTE_BIG_ENDIAN)
       xl_error ("Unsupported reference attributes in XCOM header");

     if (header.version != XCOM_VERSION)
       xl_error ("Unsupported version in XCOM header");

     sectionMap = (XComSection *) malloc (header.sections * sizeof (XComSection));
     if (sectionMap == NULL)
       goto failure;

     objectSections = (Section **) calloc (header.sections, sizeof (Section *));
     if (objectSections == NULL)
       goto failure;

     if (fread (sectionMap, header.sections * sizeof (XComSection), 1, file) != 1)
       xl_error ("Could not read section map from XCOM");
     
     for (currentSection = sectionMap;
          currentSection < & sectionMap [header.sections];
          ++ currentSection)
       if (currentSection -> type == XCOM_SECTION_TYPE_STRINGS)
         break;

     if (currentSection >= & sectionMap [header.sections])
       xl_error ("Could not find strings section in XCOM");

     stringSection = section_create (object, SECTION_TYPE_TEMPORARY, 0, 0, 0, 0, 0, 0, (size_t) currentSection -> size, NULL);
     if (stringSection == NULL)
       goto failure;

     if (fseek (file, currentSection -> offset, SEEK_SET) != 0 ||
         section_read (stringSection, file) != 0)
       xl_error ("Could not read string section from XCOM");

     for (currentSection = sectionMap;
          currentSection < & sectionMap [header.sections];
          ++ currentSection)
     {
         Section * newSection;
         SectionType type;
         const char * name = NULL,
                    * architecture = NULL;
         
         if (currentSection -> type == XCOM_SECTION_TYPE_STRINGS)
           continue;

         if (currentSection -> dataIndex != 0)
           architecture = (char *) & stringSection -> data [currentSection -> dataIndex];

         if (currentSection -> type == XCOM_SECTION_TYPE_PROGRAM &&
             strcmp (architecture, XCOM_HOST_ARCHITECTURE_CODE) != 0 &&
             strcmp (architecture, XCOM_HOST_ARCHITECTURE_DATA) != 0)
           xl_error ("Unsupported architecture type for program section in XCOM: %s",
                     architecture);

         switch (currentSection -> type)
         {
         case XCOM_SECTION_TYPE_INTERFACE:
         case XCOM_SECTION_TYPE_RESOURCE:
         case XCOM_SECTION_TYPE_PROGRAM:
                 type = (SectionType) currentSection -> type;

                 break;

         default:
                 type = SECTION_TYPE_TEMPORARY;
         }

         if (currentSection -> nameIndex != 0)
           name = (char *) & stringSection -> data [currentSection -> nameIndex];

         newSection = object_define_section (object,
                                             type,
                                             (SectionAttributes) currentSection -> sectionAttributes & SECTION_ATTRIBUTE_MASK,
                                             (SectionCapabilities) currentSection -> memoryAttributes & SECTION_CAPABILITY_MASK,
                                             name,
                                             architecture,
                                             (Address) currentSection -> memoryRegion,
                                             (size_t) currentSection -> alignment,
                                             (size_t) currentSection -> size);
       
         objectSections [currentSection - sectionMap] = newSection;

         if (currentSection -> type != XCOM_SECTION_TYPE_PROGRAM || 
             (currentSection -> sectionAttributes & XCOM_SECTION_ATTRIBUTE_CONTENTS) != 0)
         {
            if (fseek (file, currentSection -> offset, SEEK_SET) != 0 ||
                section_read (newSection, file) != 0)
              xl_error ("Failed to read section contents from XCOM");
         }
     }
     
     for (currentSection = sectionMap;
          currentSection < & sectionMap [header.sections];
          ++ currentSection)
     {
          Section * targetSection;
          XComSymbol * currentSymbol,
                     * endSymbol;

          if (currentSection -> type != XCOM_SECTION_TYPE_SYMBOLS)
            continue;

          if (currentSection -> dataIndex == 0)
            targetSection = NULL;
          else
          {
            targetSection = objectSections [currentSection -> dataIndex - 1];

            if (sectionMap [currentSection -> dataIndex - 1].type != XCOM_SECTION_TYPE_PROGRAM)
              xl_error ("Symbols exported from non-program section in XCOM");
          }

          currentSymbol = (XComSymbol *) objectSections [currentSection - sectionMap] -> data,
          endSymbol     = & currentSymbol [currentSection -> size / sizeof (XComSymbol)];
                              
          while (currentSymbol < endSymbol)
          {
              if (currentSymbol -> type != XCOM_SYMBOL_TYPE_SECTION)
                object_define_symbol (object,
                                      targetSection,
                                      (SymbolType) currentSymbol -> type,
                                      (SymbolAttributes) currentSymbol -> attributes,
                                      (char *) & stringSection -> data [currentSymbol -> nameIndex],
                                      (Address) currentSymbol -> offset,
                                      (size_t) currentSymbol -> size);

              ++ currentSymbol;
          }

          section_destroy (objectSections [currentSection - sectionMap]);
     }
                     
     for (currentSection = sectionMap;
          currentSection < & sectionMap [header.sections];
          ++ currentSection)
     {
          Section * targetSection;
          XComRelocation * currentReloc,
                         * endReloc;
          
          if (currentSection -> type != XCOM_SECTION_TYPE_RELOCATIONS)
            continue;

          if (currentSection -> dataIndex == 0 ||
              sectionMap [currentSection -> dataIndex - 1].type != XCOM_SECTION_TYPE_PROGRAM)
            xl_error ("Relocations for non-program section in XCOM");

          targetSection = objectSections [currentSection -> dataIndex - 1];
         
          currentReloc = (XComRelocation *) objectSections [currentSection - sectionMap] -> data,
          endReloc     = & currentReloc [currentSection -> size / sizeof (XComRelocation)];
                   
          while (currentReloc < endReloc)
          {
              Symbol * symbol = object_find_symbol (object, (char *) & stringSection -> data [currentReloc -> nameIndex]);

              if (symbol == NULL)
                xl_error ("Relocation against unknown symbol %s in XCOM",
                          (char *) & stringSection -> data [currentReloc -> nameIndex]);

              if (relocation_create (& targetSection -> relocations,
                                     (RelocationType) currentReloc -> relocationType,
                                     (size_t) currentReloc -> offset,
                                     symbol -> definition) == NULL)
                goto failure;

              ++ currentReloc;
          }

          section_destroy (objectSections [currentSection - sectionMap]);
     }          

     section_destroy (stringSection);
     free (sectionMap);
     free (objectSections);

     return 0;

failure:
     if (stringSection != NULL)
       section_destroy (stringSection);

     if (sectionMap != NULL)
       free (sectionMap);

     if (objectSections != NULL)
       free (objectSections);

     return -1;
}

